"""Unit tests for VendorChecksumTokenGenerator.

Version Added:
    3.0
"""

from djblets.secrets.token_generators.vendor_checksum import \
    VendorChecksumTokenGenerator
from djblets.testing.testcases import TestCase


class VendorChecksumTokenGeneratorTests(TestCase):
    """Unit tests for VendorChecksumTokenGenerator.

    Version Added:
        3.0
    """

    def setUp(self):
        super(VendorChecksumTokenGeneratorTests, self).setUp()

        self.token_generator = VendorChecksumTokenGenerator()

    def test_create_token(self):
        """Testing VendorChecksumTokenGenerator.create_token"""
        token = self.token_generator.create_token(
            token_info={
                'token_type': 'test',
            })

        self.assertEqual(len(token), 255)
        self.assertTrue(token.startswith('test'))

    def test_create_token_with_missing_required_key(self):
        """Testing VendorChecksumTokenGenerator.create_token with a missing
        required key in token info
        """
        error_message = ('The token_info dictionary must contain a '
                         'token_type key.')

        with self.assertRaisesMessage(KeyError, error_message):
            self.token_generator.create_token(token_info={})

    def test_validate_token_with_valid_token(self):
        """Testing VendorChecksumTokenGenerator.validate_token with a valid
        token
        """
        token = ('test_kEz3qSWeN25rOVA8RXitY0ywaiAOuey7WRHFeTDaqqR9g827qwuYFy5'
                 'p3NKwLThndO33aVL5pBcU5da9pNnTYtiIsC4f9uLMAP8rQKZa82W91ZArtV4'
                 'WJSDpV07VTszJq6dvzmnbvcTDQvv0crBS2XDntwx5lI4xbtpMR6mquknRneV'
                 'OJxf682I208BGhs8PdGRtv2unF176tCoC3ccI8VFx31Hdb2mgQKPyPCYEtXN'
                 '32N7RUsAX1BjxMy')

        self.assertTrue(self.token_generator.validate_token(
            token,
            token_info={'token_type': 'test'}))

    def test_validate_token_with_valid_token_wrong_base62_encode(self):
        """Testing VendorChecksumTokenGenerator.validate_token with a valid
        token that uses Djblets 3.0's incorrect base62-encoding for checksums
        """
        token = ('test_kEz3qSWeN25rOVA8RXitY0ywaiAOuey7WRHFeTDaqqR9g827qwuYFy5'
                 'p3NKwLThndO33aVL5pBcU5da9pNnTYtiIsC4f9uLMAP8rQKZa82W91ZArtV4'
                 'WJSDpV07VTszJq6dvzmnbvcTDQvv0crBS2XDntwx5lI4xbtpMR6mquknRneV'
                 'OJxf682I208BGhs8PdGRtv2unF176tCoC3ccI8VFx31Hdb2mgQKPyPCYEtXN'
                 '32N7RUsAX1bJXmY')

        self.assertTrue(self.token_generator.validate_token(
            token,
            token_info={'token_type': 'test'}))

    def test_validate_token_with_invalid_token_length(self):
        """Testing VendorChecksumTokenGenerator.validate_token with a token
        that has an invalid length
        """
        token = 'test_1234'

        self.assertFalse(self.token_generator.validate_token(
            token,
            token_info={'token_type': 'test'}))

    def test_validate_token_with_invalid_token_chars(self):
        """Testing VendorChecksumTokenGenerator.validate_token with a token
        that has an invalid character in it
        """
        token = ('test_$Ez3qSWeN25rOVA8RXitY0ywaiAOuey7WRHFeTDaqqR9g827qwuYFy5'
                 'p3NKwLThndO33aVL5pBcU5da9pNnTYtiIsC4f9uLMAP8rQKZa82W91ZArtV4'
                 'WJSDpV07VTszJq6dvzmnbvcTDQvv0crBS2XDntwx5lI4xbtpMR6mquknRneV'
                 'OJxf682I208BGhs8PdGRtv2unF176tCoC3ccI8VFx31Hdb2mgQKPyPCYEtXN'
                 '32N7RUsAX1BjxMy')

        self.assertFalse(self.token_generator.validate_token(
            token,
            token_info={'token_type': 'test'}))

    def test_validate_token_with_invalid_token_type(self):
        """Testing VendorChecksumTokenGenerator.validate_token with a token
        that has an invalid token type
        """
        token = ('fail_kEz3qSWeN25rOVA8RXitY0ywaiAOuey7WRHFeTDaqqR9g827qwuYFy5'
                 'p3NKwLThndO33aVL5pBcU5da9pNnTYtiIsC4f9uLMAP8rQKZa82W91ZArtV4'
                 'WJSDpV07VTszJq6dvzmnbvcTDQvv0crBS2XDntwx5lI4xbtpMR6mquknRneV'
                 'OJxf682I208BGhs8PdGRtv2unF176tCoC3ccI8VFx31Hdb2mgQKPyPCYEtXN'
                 '32N7RUsAX1BjxMy')

        self.assertFalse(self.token_generator.validate_token(
            token,
            token_info={'token_type': 'test'}))

    def test_validate_token_with_invalid_token_checksum(self):
        """Testing VendorChecksumTokenGenerator.validate_token with a token
        that does not match its checksum
        """
        token = ('test_kEz3qSWeN25rOVA8RXitY0ywaiAOuey7WRHFeTDaqqR9g827qwuYFy5'
                 'p3NKwLThndO33aVL5pBcU5da9pNnTYtiIsC4f9uLMAP8rQKZa82W91ZArtV4'
                 'WJSDpV07VTszJq6dvzmnbvcTDQvv0crBS2XDntwx5lI4xbtpMR6mquknRneV'
                 'OJxf682I208BGhs8PdGRtv2unF176tCoC3ccI8VFx31Hdb2mgQKPyPCYEtXN'
                 '32N7RUsAX1BjxMb')

        self.assertFalse(self.token_generator.validate_token(
            token,
            token_info={'token_type': 'test'}))

    def test_validate_token_with_empty_token_info(self):
        """Testing VendorChecksumTokenGenerator.validate_token with a valid
        token but no token_type defined in the token_info
        """
        token = ('test_kEz3qSWeN25rOVA8RXitY0ywaiAOuey7WRHFeTDaqqR9g827qwuYFy5'
                 'p3NKwLThndO33aVL5pBcU5da9pNnTYtiIsC4f9uLMAP8rQKZa82W91ZArtV4'
                 'WJSDpV07VTszJq6dvzmnbvcTDQvv0crBS2XDntwx5lI4xbtpMR6mquknRneV'
                 'OJxf682I208BGhs8PdGRtv2unF176tCoC3ccI8VFx31Hdb2mgQKPyPCYEtXN'
                 '32N7RUsAX1BjxMy')

        self.assertFalse(self.token_generator.validate_token(token,
                                                             token_info={}))



"""Testing utilities for mail-related unit tests."""

import dns.resolver
from django.core.cache import cache
from dns.rdtypes.ANY.TXT import TXT
from kgb import SpyAgency


class DmarcDnsTestsMixin(object):
    """Mixin to help with e-mail tests that need to perform DMARC lookups.

    This mixin makes it easy for unit tests to fake DMARC results, in order
    to prevent the need from looking up real data from real records.

    Unit tests can provide a mapping from domain names to TXT record strings
    in the :py:attr:`dmarc_txt_records` dictionary. When a DMARC lookup is
    performed, this dictionary will be used for the lookup.

    Note that this mixin will also clear the memory cache before each test
    run.

    Attributes:
        dmarc_txt_records (dict):
            A dictionary of domain names to DMARC TXT record strings.
    """

    def setUp(self):
        super(DmarcDnsTestsMixin, self).setUp()

        self.dmarc_txt_records = {}

        self._dmarc_spy_agency = SpyAgency()
        self._dmarc_spy_agency.spy_on(dns.resolver.resolve,
                                      call_fake=self._dns_query)
        cache.clear()

    def tearDown(self):
        super(DmarcDnsTestsMixin, self).tearDown()

        self._dmarc_spy_agency.unspy_all()

    def _dns_query(self, qname, rdtype, *args, **kwargs):
        try:
            return [TXT(1, 16, [self.dmarc_txt_records[qname]])]
        except KeyError:
            raise dns.resolver.NXDOMAIN



"""Web API resource support for OAuth2 token access."""

from django.contrib import auth
from django.core.exceptions import ImproperlyConfigured

from djblets.db.query import get_object_or_none
from djblets.webapi.errors import (LOGIN_FAILED,
                                   NOT_LOGGED_IN,
                                   OAUTH_ACCESS_DENIED_ERROR,
                                   OAUTH_MISSING_SCOPE_ERROR)


class ResourceOAuth2TokenMixin(object):
    """Augments a WebAPIResource to support OAuth2 tokens.

    Any WebAPIResource subclass making use of this mixin can accept requests
    backed by an OAuth2 token and will restrict the request to that token's
    allowed scopes.

    It is recommended that all resources in a project inherit from a base
    resource that inherits both from this mixin and from WebAPIResource.
    """

    #: Whether or not this resource is accessible when using an OAuth2 token.
    oauth2_token_access_allowed = True

    #: A mapping of HTTP methods to the type of scope required.
    #:
    #: These will be used with :py:attr:`scope_name` to generate the full
    #: required scope, e.g. :samp:`{scope_name}:read`.
    HTTP_SCOPE_METHOD_MAP = {
        'DELETE': 'destroy',
        'GET': 'read',
        'HEAD': 'read',
        'OPTIONS': 'read',
        'POST': 'write',
        'PUT': 'write',
    }

    @property
    def scope_name(self):
        """The ID used for required scopes.

        This defaults to the name of the resource, but can be overridden
        in case the name is not specific enough or there's a conflict.

        If the resource allows WebAPI tokens in addition to OAuth2 tokens, this
        will default to the resource's policy ID (which defaults to its name).
        """
        if hasattr(self, 'policy_id'):
            return self.policy_id

        return self.name

    def call_method_view(self, request, method, view, *args, **kwargs):
        """Check token scopes and call the API method handler.

        Args:
            request (django.http.HttpRequest):
                The current HTTP request.

            method (unicode):
                The HTTP method of the request.

            view (callable):
                The view function to call.

            *args (tuple):
                Additional positional arguments.

            **kwargs (dict):
                Additional keyword arguments.

        Returns:
            object:
            One of the following:

            * An error (:py:class:`djblets.webapi.errors.WebAPIError`).
            * A web API response
              (:py:class:`djblets.webapi.responces.WebAPIResponse`).
            * A tuple containing an HTTP status code and a body to serialize
              (:py:class:`tuple`).
        """
        token, from_session = self.__get_oauth2_token_for_request(request)

        if token is not None:
            if token.is_expired():
                if from_session:
                    return NOT_LOGGED_IN
                else:
                    return LOGIN_FAILED

            elif not self.oauth2_token_access_allowed:
                return OAUTH_ACCESS_DENIED_ERROR

            elif not self.__is_resource_method_allowed(method, token):
                return OAUTH_MISSING_SCOPE_ERROR

        return super(ResourceOAuth2TokenMixin, self).call_method_view(
            request, method, view, *args, **kwargs)

    def __get_oauth2_token_for_request(self, request):
        """Return the OAuth2 token for the request.

        Args:
            request (django.http.HttpRequest):
                The current HTTP request.

        Returns:
            tuple:
            A 2-tuple of the following:

            * the access token, if it exists
              (:py:class:`~oauth2_provider.models.AccessToken`); and
            * whether or not the token was retrieved from the session
              (:py:class:`bool`).
        """
        try:
            from oauth2_provider.models import AccessToken
        except ImportError:
            raise ImproperlyConfigured(
                'WebAPIOAuth2TokenAuthBackend requires django-oauth-toolkit'
                '>=0.9,<0.9.999.'
            )

        token = getattr(request, '_oauth2_token', None)
        from_session = False

        if not token:
            token_pk = request.session.get('oauth2_token_id')

            if token_pk:
                from_session = True
                token = get_object_or_none(AccessToken, pk=token_pk)

                if not token or token.is_expired():
                    # Either the token no longer exists or has expired, so we
                    # must log the user out.
                    auth.logout(request)

                request._oauth2_token = token

        return token, from_session

    def __is_resource_method_allowed(self, method, token):
        """Determine if the given method is allowed by the token.

        Args:
            method (unicode):
                The HTTP method.

            token (oauth2_provider.models.AccessToken):
                The token to check for permission.

        Returns:
            bool:
            Whether or not the token has the required scopes to perform the
            given method against this resource.
        """
        suffix = self.HTTP_SCOPE_METHOD_MAP.get(method)

        if not suffix:
            return False

        required_scope = '%s:%s' % (self.scope_name, suffix)
        return token.allow_scopes([required_scope])



"""Unit tests for classes in djblets.feedview."""

from django.test.utils import override_settings

from djblets.testing.testcases import TestCase


@override_settings(ROOT_URLCONF='djblets.feedview.test_urls')
class FeedViewTests(TestCase):
    def testViewFeedPage(self):
        """Testing view_feed with the feed-page.html template"""
        response = self.client.get('/feed/')
        self.assertContains(response, "Django 1.0 alpha released", 1)
        self.assertContains(response, "Introducing Review Board News", 1)

    def testViewFeedInline(self):
        """Testing view_feed with the feed-inline.html template"""
        response = self.client.get('/feed-inline/')
        self.assertContains(response, "Django 1.0 alpha released", 1)
        self.assertContains(response, "Introducing Review Board News", 1)

    def testViewFeedError(self):
        """Testing view_feed with a URL error"""
        response = self.client.get('/feed-error/')
        self.assertEqual(response.status_code, 200)
        self.assertTrue('error' in response.context)



"""URL patterns for applications that can use SITE_ROOT."""

from django.conf import settings
from django.conf.urls import handler404, handler500
from django.core.exceptions import ImproperlyConfigured
from django.urls import include, path


# Ensures that we can run nose on this without needing to set SITE_ROOT.
# Also serves to let people know if they set one variable without the other.
if hasattr(settings, 'SITE_ROOT'):
    if not hasattr(settings, 'SITE_ROOT_URLCONF'):
        raise ImproperlyConfigured('SITE_ROOT_URLCONF must be set when '
                                   'using SITE_ROOT')

    urlpatterns = [
        path('%s' % settings.SITE_ROOT[1:],
             include(settings.SITE_ROOT_URLCONF)),
    ]
else:
    urlpatterns = None


__all__ = [
    'handler404',
    'handler500',
    'urlpatterns',
]



import os

from django.urls import path

from djblets.feedview.views import view_feed


FEED_URL = "file://%s/testdata/sample.rss" % os.path.dirname(__file__)


urlpatterns = [
    path('feed/', view_feed, kwargs={
        'template_name': 'feedview/feed-page.html',
        'url': FEED_URL,
    }),
    path('feed-inline/', view_feed, kwargs={
        'template_name': 'feedview/feed-inline.html',
        'url': FEED_URL,
    }),
    path('feed-error/', view_feed, kwargs={
        'template_name': 'feedview/feed-inline.html',
        'url': 'http://example.fake/dummy.rss',
    }),
]



"""Unit tests for djblets.configforms.forms.ConfigPageForm."""

from django import forms
from django.contrib.auth.models import User
from django.test.client import RequestFactory

from djblets.configforms.forms import ConfigPageForm
from djblets.configforms.pages import ConfigPage
from djblets.configforms.views import ConfigPagesView
from djblets.testing.testcases import TestCase


class MyTestForm(ConfigPageForm):
    form_id = 'my-form'
    field1 = forms.CharField(label='Field 1',
                             required=False)
    field2 = forms.CharField(label='Field 2',
                             required=False)


class MyTestPage(ConfigPage):
    page_id = 'my-page'
    form_classes = [MyTestForm]


class ConfigPageFormTests(TestCase):
    """Unit tests for djblets.configforms.forms.ConfigPageForm."""

    def setUp(self):
        super(ConfigPageFormTests, self).setUp()

        request = RequestFactory().request()
        user = User.objects.create_user(username='test-user',
                                        password='test-user')
        page = MyTestPage(ConfigPagesView, request, user)

        self.form = MyTestForm(page, request, user)

    def test_initial_state(self):
        """Testing ConfigPageForm initial state"""
        self.assertEqual(self.form.fields['form_target'].initial, 'my-form')

    def test_set_initial(self):
        """Testing ConfigPageForm.set_initial"""
        self.form.set_initial({
            'field1': 'foo',
            'field2': 'bar',
        })

        self.assertEqual(self.form.fields['field1'].initial, 'foo')
        self.assertEqual(self.form.fields['field2'].initial, 'bar')

    def test_render(self):
        """Testing ConfigPageForm.render"""
        rendered = self.form.render()

        self.assertHTMLEqual(
            '<input id="id_form_target" name="form_target"'
            ' type="hidden" value="my-form">'
            '<div class="fields-row field-field1" id="row-field1">'
            ' <div class="field">'
            '  <label for="id_field1">Field 1:</label>'
            '  <input id="id_field1" name="field1" type="text">'
            ' </div>'
            '</div>'
            '<div class="fields-row field-field2" id="row-field2">'
            ' <div class="field">'
            '  <label for="id_field2">Field 2:</label>'
            '  <input id="id_field2" name="field2" type="text">'
            ' </div>'
            '</div>'
            '<input type="submit" class="btn" value="Save">',
            rendered)



"""Unit tests for djblets.privacy.consent.registry."""

from djblets.privacy.consent import BaseConsentRequirement
from djblets.privacy.consent.errors import ConsentRequirementConflictError
from djblets.privacy.consent.registry import ConsentRequirementsRegistry
from djblets.privacy.tests.testcases import ConsentTestCase


class MyConsentRequirement(BaseConsentRequirement):
    requirement_id = 'test-requirement'
    name = 'Test Requirement'
    summary = 'We want this.'
    intent_description = 'Test.'
    data_use_description = 'All.'


class ConsentRequirementsRegistryTests(ConsentTestCase):
    """Unit tests for ConsentRequirementsRegistry."""

    def setUp(self):
        super(ConsentRequirementsRegistryTests, self).setUp()

        self.registry = ConsentRequirementsRegistry()

    def test_register(self):
        """Testing ConsentRequirementsRegistry.register"""
        requirement = MyConsentRequirement()
        self.registry.register(requirement)

        self.assertEqual(list(self.registry), [requirement])

    def test_register_with_conflict(self):
        """Testing ConsentRequirementsRegistry.register with conflicting ID"""
        requirement = MyConsentRequirement()
        self.registry.register(requirement)

        with self.assertRaises(ConsentRequirementConflictError):
            self.registry.register(requirement)

    def test_get_consent_requirement(self):
        """Testing ConsentRequirementsRegistry.get_consent_requirement"""
        requirement = MyConsentRequirement()
        self.registry.register(requirement)

        self.assertEqual(
            self.registry.get_consent_requirement('test-requirement'),
            requirement)

    def test_get_consent_requirement_with_invalid_id(self):
        """Testing ConsentRequirementsRegistry.get_consent_requirement with
        invalid ID
        """
        self.assertIsNone(
            self.registry.get_consent_requirement('test-requirement'))



"""Packaging support for extensions."""

import inspect
import json
import os
import re
import sys
from distutils.errors import DistutilsExecError
from fnmatch import fnmatch

import django
import pkg_resources
from django.core.management import call_command
from setuptools.command.build_py import build_py
from setuptools import Command

from djblets.dependencies import (babel_npm_dependencies,
                                  lesscss_npm_dependencies,
                                  uglifyjs_npm_dependencies)
from djblets.util.filesystem import is_exe_in_path


class BuildStaticFiles(Command):
    """Builds static files for the extension.

    This will build the static media files used by the extension. JavaScript
    bundles will be minified and versioned. CSS bundles will be processed
    through lesscss (if using .less files), minified and versioned.

    This must be subclassed by the project offering the extension support.
    The subclass must provide the extension_entrypoint_group and
    django_settings_module parameters.

    extension_entrypoint_group is the group name that entry points register
    into.

    django_settings_module is the Python module path for the project's
    settings module, for use in the DJANGO_SETTINGS_MODULE environment
    variable.
    """

    description = 'Build static media files'
    extension_entrypoint_group = None
    django_settings_module = None

    user_options = [
        (str('remove-source-files'), None,
         'remove source files from the package'),
    ]
    boolean_options = [str('remove-source-files')]

    def initialize_options(self):
        self.build_lib = None
        self.remove_source_files = False

    def finalize_options(self):
        self.set_undefined_options('build', ('build_lib', 'build_lib'))

    def get_lessc_global_vars(self):
        """Returns a dictionary of LessCSS global variables and their values.

        This can be implemented by subclasses to provide global variables for
        .less files for processing.

        By default, this defines two variables: ``STATIC_ROOT`` and ``DEBUG``.

        ``STATIC_ROOT`` is set to ``/static/``. Any imports using
        ``@{STATIC_ROOT}`` will effectively look up the requested file in
        ``<import_path>/@{STATIC_ROOT}``. This assumes that the project
        serving the static files keeps them in :file:`static/{appname}/`.

        Projects using less.js for the runtime can then define ``STATIC_ROOT``
        to ``settings.STATIC_URL``, ensuring lookups work for development and
        packaged extensions.

        ``DEBUG`` is set to false. Runtimes using less.js can set this to
        ``settings.DEBUG`` for templates. This can be useful for LessCSS
        guards.

        This requires LessCSS 1.5.1 or higher.
        """
        return {
            'DEBUG': False,
            'STATIC_ROOT': '/static/',
        }

    def get_lessc_include_path(self):
        """Returns the include path for LessCSS imports.

        By default, this will include the parent directory of every path in
        STATICFILES_DIRS, plus the static directory of the extension.
        """
        from django.conf import settings

        less_include = set()

        for staticfile_dir in settings.STATICFILES_DIRS:
            if isinstance(staticfile_dir, tuple):
                staticfile_dir = staticfile_dir[1]

            dirname = os.path.dirname(staticfile_dir)
            less_include.add(dirname)

            if os.path.basename(dirname) == 'static':
                # This code is here for compatibility reasons. With a proper
                # Babel setup, .less files should be able to reference others
                # relative to static/ directories. However, some paths may
                # still prefix /static/ to any paths, and we don't want those
                # to break. So, add the parent of this static/ directory
                # in order to allow those lookups to continue working.
                dirname = os.path.dirname(dirname)

            less_include.add(dirname)

        return less_include

    def install_pipeline_deps(self, extension, css_bundles, js_bundles):
        """Install dependencies needed for the static media pipelining.

        This will install the LessCSS and UglifyJS tools, if needed for the
        packaging process. These will only be installed if they'd be used,
        which is determined based on the contents of the bundles.

        Subclasses can override this to support additional tools.

        Args:
            extension (djblets.extensions.extension.Extension):
                The extension being packaged.

            css_bundles (dict):
                A dictionary of CSS bundles being built for the package.

            js_bundles (dict):
                A dictionary of JavaScript bundles being built for the package.
        """
        from pipeline.conf import settings as pipeline_settings

        build_dir = os.path.join(os.getcwd(), 'build')
        node_modules_dir = os.path.join(build_dir, 'node_modules')

        os.environ['NODE_PATH'] = node_modules_dir

        if not os.path.exists(build_dir):
            os.mkdir(build_dir, 0o755)

        if not os.path.exists(node_modules_dir):
            os.mkdir(node_modules_dir, 0o755)

        dependencies = {}

        if self.get_bundle_file_matches(css_bundles, '*.less'):
            dependencies.update(lesscss_npm_dependencies)

            pipeline_settings.LESS_BINARY = \
                os.path.join(node_modules_dir, '.bin', 'lessc')

        if self.get_bundle_file_matches(js_bundles, '*.js'):
            dependencies.update(uglifyjs_npm_dependencies)

            bin_path = os.path.join(node_modules_dir, '.bin')

            pipeline_settings.UGLIFYJS_BINARY = \
                os.path.join(bin_path, 'uglifyjs')

            if self.get_bundle_file_matches(js_bundles, '*.es6.js'):
                dependencies.update(babel_npm_dependencies)

                pipeline_settings.BABEL_BINARY = \
                    os.path.join(bin_path, 'babel')

        package_file = os.path.join(build_dir, 'package.json')

        with open(package_file, 'w') as fp:
            fp.write(json.dumps(
                {
                    'name': '%s-extension' % os.path.basename(os.getcwd()),
                    'private': 'true',
                    'devDependencies': {},
                    'dependencies': dependencies,
                },
                indent=2))

        old_cwd = os.getcwd()
        os.chdir(build_dir)

        self.npm_install()

        os.chdir(old_cwd)

    def get_bundle_file_matches(self, bundles, pattern):
        """Return whether there's any files in a bundle matching a pattern.

        Args:
            bundles (dict):
                A dictionary of bundles.

            pattern (unicode):
                The filename pattern to match against.

        Returns:
            bool:
            ``True`` if a filename in one or more bundles matches the pattern.
            ``False`` if no filenames match.
        """
        for bundle_name, bundle_info in bundles.items():
            for filename in bundle_info.get('source_filenames', []):
                if fnmatch(filename, pattern):
                    return True

        return False

    def npm_install(self, package_spec=None):
        """Install a package via npm.

        This will first determine if npm is available, and then attempt to
        install the given package.

        Args:
            package_spec (unicode, optional):
                The package specification (name and optional version range)
                to install. If not specified, this will use the default
                behavior of reading :file:`package.json`.

        Raises:
            distutils.errors.DistutilsExecError:
                :command:`npm` could not be found, or there was an error
                installing the package.
        """
        if not hasattr(self, '_checked_npm'):
            if not is_exe_in_path('npm'):
                raise DistutilsExecError(
                    'Unable to locate npm in the path, which is needed to '
                    'install %s. Static media cannot be built.'
                    % package_spec)

            self._checked_npm = True

        if package_spec:
            # Ensure there's a node_modules directory here, so it doesn't
            # install higher up in the directory hierarchy.
            if not os.path.exists('node_modules'):
                os.mkdir('node_modules', 0o755)

            print('Installing %s...' % package_spec)
            result = os.system('npm install %s' % package_spec)
        else:
            print('Installing node packages...')
            result = os.system('npm install')

        if result != 0:
            raise DistutilsExecError('Installation from npm failed.')

    def run(self):
        from django.conf import settings

        # Prepare to import the project's settings file, and the extension
        # modules that are being shipped, so we can scan for the bundled
        # media.
        old_settings_module = os.environ.get('DJANGO_SETTINGS_MODULE')
        os.environ['DJANGO_SETTINGS_MODULE'] = self.django_settings_module

        # Skip any initial pipeline settings validation in the consuming
        # project's settings.py. The paths may not exist at this stage. We'll
        # be creating them.
        os.environ['DJBLETS_SKIP_PIPELINE_VALIDATION'] = '1'

        django.setup()

        cwd = os.getcwd()
        sys.path = [
            os.path.join(cwd, package_name)
            for package_name in self.distribution.packages
        ] + sys.path

        # Set up the common Django settings for the builds.
        settings.STATICFILES_FINDERS = (
            'djblets.extensions.staticfiles.PackagingFinder',
        )
        settings.STATICFILES_STORAGE = \
            'djblets.extensions.staticfiles.PackagingCachedFilesStorage'
        settings.INSTALLED_APPS = [
            'django.contrib.staticfiles',
        ]
        settings.CACHES = {
            'default': {
                'BACKEND': 'django.core.cache.backends.locmem.LocMemCache',
            },
        }

        # Load the entry points this package is providing, so we'll know
        # which extensions to scan.
        entrypoints = pkg_resources.EntryPoint.parse_map(
            self.distribution.entry_points,
            dist=self.distribution)

        extension_entrypoints = \
            entrypoints.get(self.extension_entrypoint_group)
        assert extension_entrypoints, 'No extension entry points were defined.'

        # Begin building pipeline bundles for each of the bundles defined
        # in the extension.
        for ep_name, entrypoint in extension_entrypoints.items():
            try:
                extension = entrypoint.load(require=False)
            except ImportError:
                sys.stderr.write(
                    'Error loading the extension for entry point %s\n'
                    % ep_name)
                raise

            self._build_static_media(extension)

        # Restore the environment, so we don't possibly interfere with
        # anything else.
        if old_settings_module is not None:
            os.environ['DJANGO_SETTINGS_MODULE'] = old_settings_module

        sys.path = sys.path[len(self.distribution.packages):]

    def _build_static_media(self, extension):
        from django.conf import settings
        from pipeline.conf import settings as pipeline_settings

        pipeline_js = {}
        pipeline_css = {}

        self._add_bundle(pipeline_js, extension.js_bundles, 'js', '.js')
        self._add_bundle(pipeline_css, extension.css_bundles, 'css', '.css')

        # Get the location of the static/ directory within the module in the
        # source tree. We're going to use it to look up static files for
        # input, and as a relative path within the module for the output.
        module_dir = os.path.dirname(inspect.getmodule(extension).__file__)
        static_dir = os.path.join(module_dir, 'static')

        if not os.path.exists(static_dir):
            # This extension doesn't define any static files.
            return

        # Install any dependencies that might be needed by the bundles.
        self.install_pipeline_deps(extension, pipeline_css, pipeline_js)

        from djblets.extensions.staticfiles import PackagingFinder
        PackagingFinder.extension_static_dir = static_dir

        settings.STATICFILES_DIRS = list(settings.STATICFILES_DIRS) + [
            PackagingFinder.extension_static_dir
        ]

        settings.STATIC_ROOT = \
            os.path.join(self.build_lib,
                         os.path.relpath(os.path.join(module_dir, 'static')))

        # Tell djblets.pipeline.compiles.less.LessCompiler (if enabled) not
        # to check for outdated using its special import script, and to in
        # fact assume all files are outdated so they'll all be rebuilt. This
        # will also prevent access denied issues (due to referenced files not
        # being in the sandboxed tree defined for the static media).
        pipeline_settings._DJBLETS_LESS_ALWAYS_REBUILD = True

        # Due to how Pipeline copies and stores its settings, we actually
        # have to copy over some of these, as they'll be from the original
        # loaded settings.
        from pipeline.conf import settings as pipeline_settings

        pipeline_settings.JAVASCRIPT = pipeline_js
        pipeline_settings.STYLESHEETS = pipeline_css
        pipeline_settings.PIPELINE_ENABLED = True
        pipeline_settings.LESS_ARGUMENTS = self._build_lessc_args()

        # Collect and process all static media files.
        call_command('collectstatic', interactive=False, verbosity=2)

        if self.remove_source_files:
            self._remove_source_files(
                pipeline_css, os.path.join(settings.STATIC_ROOT, 'css'))
            self._remove_source_files(
                pipeline_js, os.path.join(settings.STATIC_ROOT, 'js'))

    def _build_lessc_args(self):
        """Build the list of arguments for the less compiler.

        This will return a set of arguments that define any global variables
        or include paths for the less compiler. It respects most existing
        arguments already in ``settings.PIPELINE['LESS_ARGUMENTS']``, but
        will filter out any existing include paths and global variables needed
        for packaging.

        Returns:
            list:
            A list of arguments to pass to lessc.
        """
        from django.conf import settings

        # Get all the arguments that the application defines, but remove those
        # that we need to override or the ones that aren't compatible with
        # building this media.
        lessc_global_vars = self.get_lessc_global_vars()
        exclude_re = re.compile(
            '^--(include-path=|global-var="?(%s)=)'
            % '|'.join(
                re.escape(global_var)
                for global_var in lessc_global_vars.keys()
            ))

        lessc_args = [
            lessc_arg
            for lessc_arg in settings.PIPELINE.get('LESS_ARGUMENTS', [])
            if not exclude_re.match(lessc_arg)
        ]

        # Register the include path and any global variables used for
        # building .less files.
        return lessc_args + [
            '--include-path=%s'
            % os.path.pathsep.join(self.get_lessc_include_path())
        ] + [
            '--global-var=%s=%s'
            % (key, self._serialize_lessc_value(value))
            for key, value in lessc_global_vars.items()
        ]

    def _add_bundle(self, pipeline_bundles, extension_bundles, default_dir,
                    ext):
        for name, bundle in extension_bundles.items():
            if 'output_filename' not in bundle:
                bundle['output_filename'] = \
                    '%s/%s.min%s' % (default_dir, name, ext)

            pipeline_bundles[name] = bundle

    def _remove_source_files(self, pipeline_bundles, media_build_dir):
        """Removes all source files, leaving only built bundles."""
        for root, dirs, files in os.walk(media_build_dir, topdown=False):
            for name in files:
                # A valid file will be represented as one of:
                #
                #     (bundle_name, 'min', stamp, ext)
                #     (bundle_name, 'min', ext)
                #
                # We keep both the pre-stamped and post-stamped versions so
                # that Django's CachedFilesStorage can generate and cache
                # the stamp from the contents of the non-stamped file.
                name_parts = name.split('.')

                if (len(name_parts) < 3 or
                    name_parts[0] not in pipeline_bundles or
                    name_parts[1] != 'min'):
                    # This doesn't appear to be a file representing a bundle,
                    # so we should get rid of it.
                    os.unlink(os.path.join(root, name))

            for name in dirs:
                try:
                    os.rmdir(os.path.join(root, name))
                except Exception:
                    # The directory is probably not empty yet.
                    pass

    def _serialize_lessc_value(self, value):
        if isinstance(value, str):
            return '"%s"' % value
        elif isinstance(value, bool):
            if value:
                return 'true'
            else:
                return 'false'
        elif isinstance(value, int):
            return '%d' % value
        else:
            raise TypeError('%r is not a valid lessc global variable value'
                            % value)


class BuildPy(build_py):
    def run(self):
        self.run_command('build_static_files')
        build_py.run(self)


def build_extension_cmdclass(build_static_files_cls):
    """Builds a cmdclass to pass to setup.

    This is passed a subclass of BuildStaticFiles, and returns something
    that can be passed to setup().
    """
    return {
        'build_static_files': build_static_files_cls,
        'build_py': BuildPy,
    }



import os
import shutil
import stat
import sys
import tempfile

import nose
from django.core.management import execute_from_command_line
from django.test.runner import DiscoverRunner

try:
    # Make sure to pre-load all the image handlers. If we do this later during
    # unit tests, we don't seem to always get our list, causing tests to fail.
    from PIL import Image
    Image.init()
except ImportError:
    try:
        import Image
        Image.init()
    except ImportError:
        pass

from django.conf import settings
from djblets.cache.serials import generate_media_serial


class TestRunner(DiscoverRunner):
    """Test runner for standard Djblets-based projects.

    This class provides all the common setup for settings, databases, and
    directories that are generally needed by Django projects using Djblets.
    Much of the behavior can be overridden by subclasses.

    `nose <http://nose.readthedocs.io/en/latest/>`_ is used to run the test
    suites. The options can be configured through :py:attr:`nose_options`.

    This can be subclassed in order to set the settings for the test run, or
    it can be instantiated with those settings passed in as keyword arguments.
    """

    #: The options used for nose.
    #:
    #: This is a list of command line arguments that would be passed to
    #: :command:`nosetests`.
    nose_options = [
        '-v',
        '--match=^test',
        '--with-id',
        '--with-doctest',
        '--doctest-extension=.txt',
    ]

    #: A list of Python package/module names to test.
    test_packages = []

    #: Whether or not ``collectstatic`` needs to be run before tests.
    needs_collect_static = True

    def __init__(self, *args, **kwargs):
        """Initialize the test runner.

        The caller can override any of the options otherwise defined on the
        class.

        Args:
            nose_options (list, optional):
                A list of options used for nose. See :py:attr:`nose_options`.

            test_packages (list, optional):
                A list of Python package/module names to test. See
                :py:attr:`test_packages`.

            needs_collect_static (bool, optional):
                Whether or not ``collectstatic`` needs to be run before
                tests. See :py:attr:`needs_collect_static`.
        """
        super(TestRunner, self).__init__(*args, **kwargs)

        # Override any values that the caller wants to override. This allows
        # the runner to be instantiated with the desired arguments instead
        # of subclassed.
        try:
            self.nose_options = kwargs['nose_options']
        except KeyError:
            pass

        try:
            self.test_packages = kwargs['test_packages']
        except KeyError:
            pass

        try:
            self.needs_collect_static = kwargs['needs_collect_static']
        except KeyError:
            pass

    def setup_test_environment(self, *args, **kwargs):
        """Set up an environment for the unit tests.

        This will handle setting all the default settings for a Djblets-based
        project and will create the directory structure needed for the tests
        in a temp directory.

        Subclasses can override this to provide additional setup logic.

        This must be called before :py:meth:`run_tests`.

        Args:
            *args (tuple):
                Additional positional arguments to pass to Django's version
                of this method.

            **kwargs (dict):
                Additional keyword arguments to pass to Django's version
                of this method.
        """
        super(TestRunner, self).setup_test_environment(*args, **kwargs)

        # Default to testing in a non-subdir install.
        settings.SITE_ROOT = '/'

        # Set some defaults for cache serials, in case the tests need them.
        settings.AJAX_SERIAL = 123
        settings.TEMPLATE_SERIAL = 123

        # Set a faster password hasher, for performance.
        settings.PASSWORD_HASHERS = (
            'django.contrib.auth.hashers.SHA1PasswordHasher',
        )

        # Make sure we're using standard static files storage, and not
        # something like Pipeline or S3 (since we don't want to trigger any
        # special behavior). Subclasses are free to override this setting.
        settings.STATICFILES_STORAGE = \
            'django.contrib.staticfiles.storage.StaticFilesStorage'

        # By default, don't look up DMARC records when generating From
        # addresses for e-mails. Just assume we can, since we're not
        # sending anything out. Some unit tests will override
        # this.
        settings.EMAIL_ENABLE_SMART_SPOOFING = False

        # Create a temp directory that tests can rely upon.
        self.tempdir = tempfile.mkdtemp(prefix='rb-tests-')

        # Configure file paths for static media. This will handle the main
        # static and uploaded media directories, along with extension
        # directories (for projects that need to use them).
        settings.STATIC_URL = settings.SITE_ROOT + 'static/'
        settings.MEDIA_URL = settings.SITE_ROOT + 'media/'
        settings.STATIC_ROOT = os.path.join(self.tempdir, 'static')
        settings.MEDIA_ROOT = os.path.join(self.tempdir, 'media')

        required_dirs = self.setup_dirs() + [
            settings.STATIC_ROOT,
            settings.MEDIA_ROOT,
            os.path.join(settings.MEDIA_ROOT, 'ext'),
            os.path.join(settings.STATIC_ROOT, 'ext'),
        ]

        for dirname in required_dirs:
            if not os.path.exists(dirname):
                os.makedirs(dirname)

        if self.needs_collect_static:
            # Collect all static media needed for tests.
            execute_from_command_line([
                __file__, 'collectstatic', '--noinput', '-v', '0',
            ])

        generate_media_serial()

    def teardown_test_environment(self, *args, **kwargs):
        """Tear down the environment for the unit tests.

        This will clean up the temp directory structure.It must be called after
        :py:meth:`run_tests`.

        Args:
            *args (tuple):
                Additional positional arguments to pass to Django's version
                of this method.

            **kwargs (dict):
                Additional keyword arguments to pass to Django's version
                of this method.
        """
        shutil.rmtree(self.tempdir)

        super(TestRunner, self).teardown_test_environment(*args, **kwargs)

    def run_tests(self, test_labels=[], argv=None, *args, **kwargs):
        """Run the test suite.

        Args:
            test_labels (list of unicode, optional):
                Specific tests to run.

            argv (list of unicode, optional):
                Additional arguments for nose. If not specified, sys.argv is
                used.

            *args (tuple, unused):
                Unused additional positional arguments.

            **kwargs (dict, unused):
                Unused additional keyword arguments.

        Returns:
            int:
            The exit code. 0 means all tests passed, while 1 means there were
            failures.
        """
        if argv is None:
            argv = sys.argv

        self.setup_test_environment()
        old_config = self.setup_databases()

        self.nose_argv = [argv[0]] + self.nose_options

        if '--with-coverage' in argv:
            self.nose_argv += ['--with-coverage'] + [
                '--cover-package=%s' % package_name
                for package_name in self.test_packages
            ]
            argv.remove('--with-coverage')

        # If the test files are executable on the file system, nose will need
        # the --exe argument to run them
        known_file = os.path.join(os.path.dirname(__file__), '__init__.py')

        if (os.path.exists(known_file) and
            os.stat(known_file).st_mode & stat.S_IXUSR):
            self.nose_argv.append('--exe')

        # manage.py captures everything before "--"
        if len(argv) > 2 and '--' in argv:
            self.nose_argv += argv[(argv.index('--') + 1):]

        # test_labels may be provided to us with some command line arguments,
        # which we would have already added above. We need to sanitize this
        # when adding it to the argument list.
        self.nose_argv += [
            test_label
            for test_label in test_labels
            if (not test_label.startswith('-') and
                test_label not in self.nose_argv)
        ]

        # If specific tests are not requested, test all the configured
        # test packages.
        specific_tests = [
            test_name
            for test_name in self.nose_argv[1:]
            if not test_name.startswith('-')
        ]

        if not specific_tests:
            self.nose_argv += self.test_packages

        self.run_nose()

        self.teardown_databases(old_config)
        self.teardown_test_environment()

        if self.result.success:
            return 0
        else:
            return 1

    def setup_dirs(self):
        """Set up directories to create and use.

        This can return one or more directory paths that need to be created
        before the tests can be run. It may also store settings pointing to
        those paths.

        This is not responsible for creating the directories. Any returned
        paths will be created automatically.

        Returns:
            list of unicode:
            A list of directory paths to create.
        """
        return []

    def run_nose(self):
        """Run the unit tests using nose.

        This will use nose to run the tests, storing the result.

        Returns:
            nose.core.TestProgram:
            The result from the run.
        """
        self.result = nose.main(argv=self.nose_argv, exit=False)



"""Built-in resource representing the User model."""

from django.contrib.auth.models import User

from djblets.util.decorators import augment_method_from
from djblets.webapi.fields import IntFieldType, StringFieldType
from djblets.webapi.resources.base import WebAPIResource
from djblets.webapi.resources.registry import register_resource_for_model


class UserResource(WebAPIResource):
    """A default resource for representing a Django User model."""

    model = User
    fields = {
        'id': {
            'type': IntFieldType,
            'description': 'The numeric ID of the user.',
        },
        'username': {
            'type': StringFieldType,
            'description': "The user's username.",
        },
        'first_name': {
            'type': StringFieldType,
            'description': "The user's first name.",
        },
        'last_name': {
            'type': StringFieldType,
            'description': "The user's last name.",
        },
        'fullname': {
            'type': StringFieldType,
            'description': "The user's full name (first and last).",
        },
        'email': {
            'type': StringFieldType,
            'description': "The user's e-mail address",
        },
        'url': {
            'type': StringFieldType,
            'description': "The URL to the user's page on the site. "
                           "This is deprecated and will be removed in a "
                           "future version.",
        },
    }

    uri_object_key = 'username'
    uri_object_key_regex = r'[A-Za-z0-9@\._\-\'\+]+'
    model_object_key = 'username'
    autogenerate_etags = True

    allowed_methods = ('GET',)

    def serialize_fullname_field(self, user, **kwargs):
        return user.get_full_name()

    def serialize_url_field(self, user, **kwargs):
        return user.get_absolute_url()

    def has_modify_permissions(self, request, user, *args, **kwargs):
        """Return whether or not the user can modify this object."""
        return request.user.is_authenticated and user.pk == request.user.pk

    @augment_method_from(WebAPIResource)
    def get_list(self, *args, **kwargs):
        """Retrieve the list of users on the site."""
        pass


user_resource = UserResource()
register_resource_for_model(User, user_resource)



"""Base support for creating service integrations."""

from djblets.integrations.forms import IntegrationConfigForm


class Integration(object):
    """Base class for an integration.

    Integrations are pluggable components that interface the application with
    a third-party service, notifying the service or triggering actions on
    certain events, or fetching data from the service. They can be registered
    by the consuming application or through extensions.

    Unlike an extension, an integration can contain multiple configurations
    active at one time. An example would be when you want multiple, distinct
    configurations for posting messages to different channels on a chat
    service.

    There's one Integration instance for each class, and it typically operates
    by responding to events and communicating with another service, making use
    of the state stored in one or more
    :py:class:`~djblets.integrations.models.BaseIntegrationConfig` instances,
    which it can query. This allows hook registration and other logic to be
    shared across all configurations of an instance.

    Integrations can make use of extension hooks, binding the lifecycle of that
    hook's registration to the lifecycle of that particular integration, making
    it very easy to tie an integration into any part of the application.
    """

    #: The unique identifier of the integration.
    #:
    #: This identifier cannot collide with other integrations, and thus should
    #: contain some unique information, such as a vendor name.
    #:
    #: If not set, this will default to the full class path for the
    #: integration subclass.
    integration_id = None

    #: The display name of the integration.
    name = None

    #: A short description of this integration, in plain text format.
    description = None

    #: Static paths for the integration's icon.
    #:
    #: This is a dictionary that maps icon resolution indicators to URLs or
    #: relative static media paths. It must contain a ``1x`` key, and can
    #: contain ``2x``, ``3x``, and other sizes. For now, only ``1x`` and
    #: ``2x`` are used.
    #:
    #: Each value must be a URL reachable by the browser, whether an absolute
    #: URL or a path relative to the root of the web server.
    #:
    #: The base ``1x`` icon itself is expected to be 48x48. The other sizes
    #: are expected to be multiples of this size.
    #:
    #: If the paths need to be computed (through a call to
    #: :py:func:`~django.templatetags.static` or
    #: :py:meth:`~djblets.extensions.extension.Extension.get_static_url`,
    #: it is best to declare this as a method and use
    #: :py:func:`~django.utils.functional.cached_property`.
    #:
    #: Example:
    #:     .. code-block:: python
    #:
    #:        def get_icon_urls(self):
    #:            return {
    #:                '1x': self.extension.get_static_url('logo.png'),
    #:                '2x': self.extension.get_static_url('logo@2x.png'),
    #:            }
    icon_static_urls = {}

    #: Default settings for any configurations on this extension.
    #:
    #: If a setting hasn't been explicitly saved in a configuration, it will
    #: use the default from here, if available.
    default_settings = {}

    #: The form class for handling integration configuration.
    config_form_cls = IntegrationConfigForm

    #: The template name used for the configuration page.
    config_template_name = None

    def __init__(self, integration_mgr):
        """Construct the integration.

        Implementations of an integration should generally not override this.
        Instead, they should implement :py:meth:`initialize`.

        Args:
            integration_mgr (djblets.integrations.manager.IntegrationManager):
                The integration manager that manages this integration.
        """
        self.integration_mgr = integration_mgr
        self.hooks = set()
        self.enabled = False

    @property
    def id(self):
        """The ID of the integration.

        This is an alias around :py:attr:`integration_id`, meant to provide
        compatibility with an extension's ``id`` attribute.
        """
        return self.integration_id

    def enable_integration(self):
        """Enable this integration.

        This will initialize the integration, if not already enabled,
        allowing it to listen to signals and register hooks.
        """
        if not self.enabled:
            self.enabled = True
            self.initialize()

    def disable_integration(self):
        """Disable this integration.

        This will shut down the integration, if not already disabled,
        cleaning up any signal handlers or hooks it registered.
        """
        if self.enabled:
            self.enabled = False
            self.shutdown()
            self.shutdown_hooks()
            self.hooks = set()

    def initialize(self):
        """Initialize the integration.

        Integration implementation subclasses must override this to provide
        any initialization needed for the integration, including signal and
        hook registration.

        This should only be called from :py:meth:`enable_integration`.
        """
        raise NotImplementedError('%r must implement initialize()'
                                  % self.__class__)

    def shutdown(self):
        """Shut down the integration.

        Integration implementation subclasses can override this to perform
        any cleanup work for the integration. It will be called when shutting
        down the integration.

        By default, this doesn't do anything. Subclasses do not need to
        call the parent method.
        """
        pass

    def shutdown_hooks(self):
        """Shut down all the hooks for the integration.

        By default, this is called when calling :py:meth:`shutdown`.
        """
        for hook in self.hooks.copy():
            if hook.initialized:
                hook.disable_hook()

    def get_configs(self, **filter_kwargs):
        """Return configurations matching the given filters.

        By default, all enabled configurations will be returned for this
        integration. This can be filtered down by specifying fields in the
        model and their values through keyword arguments.

        Each set of results for a unique combination of integration class and
        filter arguments will be cached locally, to speed up further lookups.

        Args:
            **filter_kwargs (dict, optional):
                Keyword arguments to filter by. Each must match a field and
                value on the model.

        Returns:
            list of djblets.integrations.models.BaseIntegrationConfig:
            A list of enabled integration configurations matching the query.
        """
        return self.integration_mgr.get_integration_configs(
            integration_cls=self.__class__,
            enabled=True,
            **filter_kwargs)

    def create_config(self, save=False, **kwargs):
        """Create a configuration for this integration.

        This will create a brand new configuration for this integration,
        with fields populated by the provided keyword arguments. The object
        created will be based on the
        :py:attr:`djblets.integrations.manager.IntegrationManager.config_model`
        attribute.

        By default, this configuration won't be automatically saved to the
        database. The caller can pass ``save=True`` to change this behavior.
        Otherwise, it's the caller's responsibility to save the configuration.

        Subclasses can override this to provide additional defaults or state.

        Args:
            save (bool, optional):
                Whether to immediately save the new configuration. Defaults
                to ``False``.

            **kwargs (dict):
                Values for fields on the configuration model.

        Returns:
            djblets.integrations.models.BaseIntegrationConfig:
            The new integration configuration instance.
        """
        config = self.integration_mgr.config_model(
            integration_id=self.integration_id,
            **kwargs)

        if save:
            config.save()

        return config



"""View mixins and decorators for requiring consent."""

from functools import wraps

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.http import HttpResponseRedirect
from django.utils.decorators import method_decorator

from djblets.privacy.consent import (Consent,
                                     get_consent_requirements_registry,
                                     get_consent_tracker)
from djblets.privacy.consent.common import PolicyConsentRequirement


_CONSENT_REDIRECT_SETTING = 'DJBLETS_PRIVACY_PENDING_CONSENT_REDIRECT_URL'


def check_pending_consent(view):
    """A decorator for ensuring the user has no pending consent requirements.

    If the user does, they will be redirected to
    ``settings.DJBLETS_PRIVACY_PENDING_CONSENT_REDIRECT_URL``.

    Args:
        view (callable):
            The view to decorate

    Returns:
        callable:
        The decorated view.
    """
    @wraps(view)
    def decorated(request, *args, **kwargs):
        user = request.user

        if user.is_authenticated:
            pending_requirements = \
                get_consent_tracker().get_pending_consent_requirements(user)
            policy_requirement = \
                get_consent_requirements_registry().get_consent_requirement(
                    PolicyConsentRequirement.requirement_id)

            if (pending_requirements or
                (policy_requirement is not None and
                 (policy_requirement.get_consent(user) !=
                  Consent.GRANTED))):
                redirect_url = getattr(settings, _CONSENT_REDIRECT_SETTING,
                                       None)

                if redirect_url is None:
                    raise ImproperlyConfigured(
                        'settings.%s must be set.' % _CONSENT_REDIRECT_SETTING
                    )

                if callable(redirect_url):
                    redirect_url = redirect_url(request)

                return HttpResponseRedirect(redirect_url)

        return view(request, *args, **kwargs)

    return decorated


class CheckPendingConsentMixin(object):
    """A view mixin for ensuring the user has no pending consent requirements.

    If the user does, they will be redirected to
    ``settings.DJBLETS_PRIVACY_PENDING_CONSENT_REDIRECT_URL``

    This mixin requires the use of
    :py:class:`~djblets.views.generic.base.PrePostDispatchViewMixin`.
    """

    @method_decorator(check_pending_consent)
    def pre_dispatch(self, *args, **kwargs):
        """Dispatch the request.

        Args:
            *args (tuple):
                Positional arguments from the URL resolver.

            **kwargs (dict):
                Keyword arguments from the URL resolver.

        Returns:
            django.http.HttpResponse:
            Either a redirect or ``None``.
        """
        return super(CheckPendingConsentMixin, self).pre_dispatch(
            *args, **kwargs)



"""Testing feature-related decorators."""

from django.http import (HttpResponse, HttpResponseForbidden,
                         HttpResponseNotFound)
from kgb import SpyAgency

from djblets.features.decorators import feature_required
from djblets.features.testing import override_feature_check
from djblets.features.tests.test_feature import DummyFeature
from djblets.features import get_features_registry
from djblets.features.checkers import set_feature_checker
from djblets.testing.testcases import TestCase


class FeatureRequiredTests(SpyAgency, TestCase):
    """Tests for the @feature_required decorator."""

    @classmethod
    def setUpClass(cls):
        super(FeatureRequiredTests, cls).setUpClass()

        cls.features = get_features_registry()

    def tearDown(self):
        super(FeatureRequiredTests, self).tearDown()

        self.features.reset()
        set_feature_checker(None)

    def test_feature_required_disabled(self):
        """Testing @feature_required() with a disabled feature"""
        feature = DummyFeature()
        self.assertFalse(feature.is_enabled())

        @feature_required(feature)
        def view(request):
            assert False, "view() should not be called."

        result = view(None)
        self.assertIsInstance(result, HttpResponseNotFound)

    def test_feature_require_enabled(self):
        """Testing @feature_required with an enabled feature"""
        feature = DummyFeature()
        self.assertFalse(feature.is_enabled())

        def view(request):
            return HttpResponse(content=b'foo')

        decorated = feature_required(feature)(view)

        self.spy_on(view)

        with override_feature_check(feature.feature_id, enabled=True):
            self.assertTrue(feature.is_enabled())

            result = decorated(None)
            self.assertIsInstance(result, HttpResponse)
            self.assertEqual(result.content, b'foo')
            self.assertTrue(view.spy.called)

    def test_feature_required_disabled_backup_view(self):
        """Testing @feature_required with a disabled view calls the correct
        backup view
        """
        feature = DummyFeature()
        self.assertFalse(feature.is_enabled())

        def _403(*args, **kwargs):
            return HttpResponseForbidden()

        @feature_required(feature, _403)
        def view(request):
            assert False, "view() should not be called."

        result = view(None)
        self.assertIsInstance(result, HttpResponseForbidden)



"""Unit tests for djblets.webapi.fields."""

from datetime import datetime

import pytz
from django.core.exceptions import ValidationError
from django.core.files.uploadedfile import UploadedFile
from django.utils import timezone
from kgb import SpyAgency

from djblets.testing.testcases import TestCase
from djblets.webapi.fields import (BooleanFieldType,
                                   ChoiceFieldType,
                                   DateTimeFieldType,
                                   DictFieldType,
                                   FileFieldType,
                                   IntFieldType,
                                   ListFieldType,
                                   ResourceFieldType,
                                   ResourceListFieldType,
                                   StringFieldType)
from djblets.webapi.resources.user import UserResource, user_resource


class BooleanFieldTypeTests(TestCase):
    """Unit tests for djblets.webapi.fields.BooleanFieldType."""

    def setUp(self):
        super(BooleanFieldTypeTests, self).setUp()

        self.field_type = BooleanFieldType({})

    def test_get_value_from_data(self):
        """Testing BooleanFieldType.get_value_from_data"""
        value = self.field_type.get_value_from_data(
            name='var',
            fields_data={
                'var': True,
            },
            files_data={
                'var': False,
            })
        self.assertTrue(value)

    def test_clean_value_with_truthy_values(self):
        """Testing BooleanFieldType.clean_value with truthy values"""
        self.assertTrue(self.field_type.clean_value(True))
        self.assertTrue(self.field_type.clean_value(1))
        self.assertTrue(self.field_type.clean_value('1'))
        self.assertTrue(self.field_type.clean_value('true'))
        self.assertTrue(self.field_type.clean_value('True'))
        self.assertTrue(self.field_type.clean_value('TRUE'))

    def test_clean_value_with_falsy_values(self):
        """Testing BooleanFieldType.clean_value with falsy values"""
        self.assertFalse(self.field_type.clean_value(False))
        self.assertFalse(self.field_type.clean_value(0))
        self.assertFalse(self.field_type.clean_value(''))
        self.assertFalse(self.field_type.clean_value('false'))
        self.assertFalse(self.field_type.clean_value('False'))
        self.assertFalse(self.field_type.clean_value('FALSE'))

    def test_str(self):
        """Testing BooleanFieldType.__str__"""
        self.assertEqual(str(self.field_type), 'Boolean')


class ChoiceFieldTypeTests(TestCase):
    """Unit tests for djblets.webapi.fields.ChoiceFieldType."""

    def setUp(self):
        super(ChoiceFieldTypeTests, self).setUp()

        self.field_type = ChoiceFieldType({
            'choices': ['a', 'b', 'cde'],
        })

    def test_init_without_choices(self):
        """Testing ChoiceFieldType.__init__ without choices key"""
        expected_message = 'Missing "choices" key in {}'

        with self.assertRaisesMessage(KeyError, expected_message):
            ChoiceFieldType({})

    def test_get_value_from_data(self):
        """Testing ChoiceFieldType.get_value_from_data"""
        value = self.field_type.get_value_from_data(
            name='var',
            fields_data={
                'var': 'b',
            },
            files_data={
                'var': 'cde',
            })
        self.assertEqual(value, 'b')

    def test_clean_value_with_valid_values(self):
        """Testing ChoiceFieldType.clean_value with valid values"""
        self.assertEqual(self.field_type.clean_value('a'), 'a')
        self.assertEqual(self.field_type.clean_value('b'), 'b')
        self.assertEqual(self.field_type.clean_value('cde'), 'cde')

    def test_clean_value_with_invalid_values(self):
        """Testing ChoiceFieldType.clean_value with invalid values"""
        expected_message = \
            '"c" is not a valid value. Valid values are: "a", "b", "cde"'

        with self.assertRaisesMessage(ValidationError, expected_message):
            self.field_type.clean_value('c')

    def test_str(self):
        """Testing ChoiceFieldType.__str__"""
        self.assertEqual(str(self.field_type),
                         'One of "a", "b", "cde"')


class DateTimeFieldTypeTests(SpyAgency, TestCase):
    """Unit tests for djblets.webapi.fields.DateTimeFieldType."""

    def setUp(self):
        super(DateTimeFieldTypeTests, self).setUp()

        self.field_type = DateTimeFieldType({})

    def test_get_value_from_data(self):
        """Testing DateTimeFieldType.get_value_from_data"""
        value = self.field_type.get_value_from_data(
            name='var',
            fields_data={
                'var': '2018-02-20T13:42:00',
            },
            files_data={
                'var': None,
            })
        self.assertEqual(value, '2018-02-20T13:42:00')

    def test_clean_value_with_aware_datetime(self):
        """Testing DateTimeFieldType.clean_value with aware datetime"""
        dt = datetime(2018, 2, 20, 13, 42, 0, tzinfo=timezone.utc)

        self.assertEqual(self.field_type.clean_value(dt),
                         dt)

    def test_clean_value_with_naive_datetime(self):
        """Testing DateTimeFieldType.clean_value with naive datetime"""
        dt = datetime(2018, 2, 20, 13, 42, 0)
        self.assertTrue(timezone.is_naive(dt))

        pst = pytz.timezone('US/Pacific')

        with timezone.override(pst):
            self.assertEqual(self.field_type.clean_value(dt),
                             pst.localize(dt))

    def test_clean_value_with_timestamp_string_utc(self):
        """Testing DateTimeFieldType.clean_value with timestamp string with
        UTC offset
        """
        self.assertEqual(
            self.field_type.clean_value('2018-02-20T13:42:00Z'),
            datetime(2018, 2, 20, 13, 42, 0, tzinfo=timezone.utc))

    def test_clean_value_with_timestamp_string_no_timezone(self):
        """Testing DateTimeFieldType.clean_value with timestamp string without
        timezone offset
        """
        pst = pytz.timezone('US/Pacific')

        with timezone.override(pst):
            self.assertEqual(
                self.field_type.clean_value('2018-02-20T13:42:00'),
                pst.localize(datetime(2018, 2, 20, 13, 42, 0)))

    def test_clean_value_with_timestamp_string_ambiguous(self):
        """Testing DateTimeFieldType.clean_value with timestamp string with
        ambiguous time
        """
        with timezone.override(pytz.timezone('America/Chicago')):
            expected_message = (
                'This timestamp needs a UTC offset to avoid being ambiguous '
                'due to daylight savings time changes'
            )

            with self.assertRaisesMessage(ValidationError, expected_message):
                self.field_type.clean_value('2016-11-06T01:05:59')

    def test_clean_value_with_invalid_timestamp(self):
        """Testing DateTimeFieldType.clean_value with invalid timestamp string
        """
        expected_message = 'This timestamp is not a valid ISO 8601 date/time'

        with self.assertRaisesMessage(ValidationError, expected_message):
            self.field_type.clean_value('2016-112')

    def test_str(self):
        """Testing DateTimeFieldType.__str__"""
        self.assertEqual(str(self.field_type),
                         'ISO 8601 Date/Time')


class DictFieldTypeTests(TestCase):
    """Unit tests for djblets.webapi.fields.DictFieldType."""

    def setUp(self):
        super(DictFieldTypeTests, self).setUp()

        self.field_type = DictFieldType({})

    def test_get_value_from_data(self):
        """Testing DictFieldType.get_value_from_data"""
        value = self.field_type.get_value_from_data(
            name='var',
            fields_data={
                'var': '{"a": true}',
            },
            files_data={
                'var': None,
            })
        self.assertEqual(value, '{"a": true}')

    def test_clean_value_with_dict(self):
        """Testing DictFieldType.clean_value with dict value"""
        self.assertEqual(
            self.field_type.clean_value({
                'a': True,
                'b': False,
            }),
            {
                'a': True,
                'b': False,
            })

    def test_clean_value_with_dict_json(self):
        """Testing DictFieldType.clean_value with JSON dict value"""
        self.assertEqual(
            self.field_type.clean_value('{"a": true, "b": false}'),
            {
                'a': True,
                'b': False,
            })

    def test_clean_value_with_non_dict(self):
        """Testing DictFieldType.clean_value with non-dict value"""
        expected_message = 'This value is not a valid dictionary value'

        with self.assertRaisesMessage(ValidationError, expected_message):
            self.field_type.clean_value([1, 2, 3])

    def test_clean_value_with_non_dict_json(self):
        """Testing DictFieldType.clean_value with JSON non-dict value"""
        expected_message = 'This value is not a valid dictionary value'

        with self.assertRaisesMessage(ValidationError, expected_message):
            self.field_type.clean_value('[1, 2, 3]')

    def test_clean_value_with_invalid_json(self):
        """Testing DictFieldType.clean_value with invalid JSON value"""
        expected_message = 'This value is not a valid JSON document'

        with self.assertRaisesMessage(ValidationError, expected_message):
            self.field_type.clean_value('{"a": 1')

    def test_str(self):
        """Testing DictFieldType.__str__"""
        self.assertEqual(str(self.field_type), 'Dictionary')


class FileFieldTypeTests(TestCase):
    """Unit tests for djblets.webapi.fields.FileFieldType."""

    def setUp(self):
        super(FileFieldTypeTests, self).setUp()

        self.field_type = FileFieldType({})

    def test_get_value_from_data(self):
        """Testing FileFieldType.get_value_from_data"""
        value = self.field_type.get_value_from_data(
            name='var',
            fields_data={
                'var': 'field',
            },
            files_data={
                'var': 'file',
            })
        self.assertEqual(value, 'file')

    def test_clean_value(self):
        """Testing FileFieldType.clean_value"""
        uploaded_file = UploadedFile()
        self.assertEqual(self.field_type.clean_value(uploaded_file),
                         uploaded_file)

    def test_str(self):
        """Testing FileFieldType.__str__"""
        self.assertEqual(str(self.field_type), 'Uploaded file')


class IntFieldTypeTests(TestCase):
    """Unit tests for djblets.webapi.fields.IntFieldType."""

    def setUp(self):
        super(IntFieldTypeTests, self).setUp()

        self.field_type = IntFieldType({})

    def test_get_value_from_data(self):
        """Testing IntFieldType.get_value_from_data"""
        value = self.field_type.get_value_from_data(
            name='var',
            fields_data={
                'var': '10',
            },
            files_data={
                'var': '20',
            })
        self.assertEqual(value, '10')

    def test_clean_value_with_valid_values(self):
        """Testing IntFieldType.clean_value with valid values"""
        self.assertEqual(self.field_type.clean_value(100), 100)
        self.assertEqual(self.field_type.clean_value('100'), 100)

    def test_clean_value_with_invalid_values(self):
        """Testing IntFieldType.clean_value with invalid values"""
        expected_message = '"c" is not an integer'

        with self.assertRaisesMessage(ValidationError, expected_message):
            self.field_type.clean_value('c')

    def test_str(self):
        """Testing IntFieldType.__str__"""
        self.assertEqual(str(self.field_type), 'Integer')


class ListFieldTypeTests(TestCase):
    """Unit tests for djblets.webapi.fields.ListFieldType."""

    def setUp(self):
        super(ListFieldTypeTests, self).setUp()

        self.field_type = ListFieldType({})

    def test_get_value_from_data(self):
        """Testing ListFieldType.get_value_from_data"""
        value = self.field_type.get_value_from_data(
            name='var',
            fields_data={
                'var': '[1, 2, 3]',
            },
            files_data={
                'var': None,
            })
        self.assertEqual(value, '[1, 2, 3]')

    def test_clean_value_with_list(self):
        """Testing ListFieldType.clean_value with list value"""
        self.assertEqual(self.field_type.clean_value([1, 2, 3]),
                         [1, 2, 3])

    def test_clean_value_with_list_json(self):
        """Testing ListFieldType.clean_value with JSON list value"""
        self.assertEqual(self.field_type.clean_value('[1, 2, 3]'),
                         [1, 2, 3])

    def test_clean_value_with_list_and_item_type_and_valid_item(self):
        """Testing ListFieldType.clean_value with list value and item type with
        valid items
        """
        field_type = ListFieldType({
            'items': {
                'type': BooleanFieldType,
            },
        })
        self.assertEqual(field_type.clean_value(['1', 'true', False]),
                         [True, True, False])

    def test_clean_value_with_list_and_item_type_and_invalid_item(self):
        """Testing ListFieldType.clean_value with list value and item type with
        invalid item
        """
        field_type = ListFieldType({
            'items': {
                'type': ChoiceFieldType,
                'choices': ['a', 'b'],
            },
        })

        expected_message = \
            '"c" is not a valid value. Valid values are: "a", "b"'

        with self.assertRaisesMessage(ValidationError, expected_message):
            field_type.clean_value(['a', 'b', 'c'])

    def test_clean_value_with_non_list(self):
        """Testing ListFieldType.clean_value with non-list value"""
        expected_message = 'This value is not a valid list'

        with self.assertRaisesMessage(ValidationError, expected_message):
            self.field_type.clean_value(None)

    def test_clean_value_with_non_list_json(self):
        """Testing ListFieldType.clean_value with non-list value"""
        expected_message = 'This value is not a valid list'

        with self.assertRaisesMessage(ValidationError, expected_message):
            self.field_type.clean_value('{}')

    def test_clean_value_with_invalid_json(self):
        """Testing ListFieldType.clean_value with invalid JSON value"""
        expected_message = 'This value is not a valid JSON document'

        with self.assertRaisesMessage(ValidationError, expected_message):
            self.field_type.clean_value('[1, 2, 3')

    def test_str(self):
        """Testing ListFieldType.__str__"""
        self.assertEqual(str(self.field_type), 'List')

    def test_str_with_item_type(self):
        """Testing ListFieldType.__str__ with item type"""
        field_type = ListFieldType({
            'items': {
                'type': ChoiceFieldType,
                'choices': ['a', 'b'],
            },
        })

        self.assertEqual(str(field_type), 'List of One of "a", "b"')


class ResourceFieldTypeTests(TestCase):
    """Unit tests for djblets.webapi.fields.ResourceFieldType."""

    def setUp(self):
        super(ResourceFieldTypeTests, self).setUp()

        self.field_type = ResourceFieldType({
            'resource': user_resource,
        })

    def test_init_without_resource(self):
        """Testing ResourceFieldType.__init__ without resource key"""
        expected_message = 'Missing "resource" key in {}'

        with self.assertRaisesMessage(KeyError, expected_message):
            ResourceFieldType({})

    def test_init_with_resource_instance(self):
        """Testing ResourceFieldType.__init__ with resource instance"""
        self.assertEqual(self.field_type.resource, user_resource)

    def test_init_with_resource_class(self):
        """Testing ResourceFieldType.__init__ with resource class"""
        field_type = ResourceFieldType({
            'resource': UserResource,
        })
        self.assertEqual(field_type.resource, UserResource)

    def test_init_with_resource_instance_path_string(self):
        """Testing ResourceFieldType.__init__ with resource instance path
        string
        """
        field_type = ResourceFieldType({
            'resource': 'djblets.webapi.resources.user.user_resource',
        })
        self.assertEqual(field_type.resource, user_resource)

    def test_init_with_resource_class_path_string(self):
        """Testing ResourceFieldType.__init__ with resource class path string
        """
        field_type = ResourceFieldType({
            'resource': 'djblets.webapi.resources.user.UserResource',
        })
        self.assertEqual(field_type.resource, UserResource)

    def test_init_with_resource_path_string_and_missing(self):
        """Testing ResourceFieldType.__init__ with resource path string and
        resource not found
        """
        # Note that we're not checking the entirety of the string here,
        # since Python 2 and 3 differ in how they represent the module that
        # failed to import.
        expected_error = (
            'Unable to load resource "foo.bar.my_resource": No module '
            'named'
        )

        with self.assertRaisesMessage(ImportError, expected_error):
            ResourceFieldType({
                'resource': 'foo.bar.my_resource',
            })

    def test_get_value_from_data(self):
        """Testing ResourceFieldType.get_value_from_data"""
        value = self.field_type.get_value_from_data(
            name='var',
            fields_data={
                'var': 'abc',
            },
            files_data={
                'var': 'def',
            })
        self.assertEqual(value, 'abc')

    def test_clean_value(self):
        """Testing ResourceFieldType.clean_value not supported"""
        expected_message = \
            'ResourceFieldType cannot be used for request data'

        with self.assertRaisesMessage(NotImplementedError, expected_message):
            self.field_type.clean_value('abc')

    def test_str(self):
        """Testing ResourceFieldType.__str__"""
        self.assertEqual(str(self.field_type), 'UserResource')


class ResourceListFieldTypeTests(TestCase):
    """Unit tests for djblets.webapi.fields.ResourceListFieldType."""

    def setUp(self):
        super(ResourceListFieldTypeTests, self).setUp()

        self.field_type = ResourceListFieldType({
            'resource': user_resource,
        })

    def test_init_without_resource(self):
        """Testing ResourceListFieldType.__init__ without resource key"""
        expected_message = 'Missing "resource" key in {}'

        with self.assertRaisesMessage(KeyError, expected_message):
            ResourceListFieldType({})

    def test_init_with_resource_instance(self):
        """Testing ResourceListFieldType.__init__ with resource instance"""
        self.assertEqual(self.field_type.resource, user_resource)

    def test_init_with_resource_path_string(self):
        """Testing ResourceListFieldType.__init__ with resource path string"""
        field_type = ResourceListFieldType({
            'resource': 'djblets.webapi.resources.user.user_resource',
        })
        self.assertEqual(field_type.resource, user_resource)

    def test_init_with_resource_path_string_and_missing(self):
        """Testing ResourceListFieldType.__init__ with resource path string and
        resource not found
        """
        # Note that we're not checking the entirety of the string here,
        # since Python 2 and 3 differ in how they represent the module that
        # failed to import.
        expected_error = (
            'Unable to load resource "foo.bar.my_resource": No module '
            'named'
        )

        with self.assertRaisesMessage(ImportError, expected_error):
            ResourceListFieldType({
                'resource': 'foo.bar.my_resource',
            })

    def test_get_value_from_data(self):
        """Testing ResourceListFieldType.get_value_from_data"""
        value = self.field_type.get_value_from_data(
            name='var',
            fields_data={
                'var': 'abc',
            },
            files_data={
                'var': 'def',
            })
        self.assertEqual(value, 'abc')

    def test_clean_value(self):
        """Testing ResourceListFieldType.clean_value not supported"""
        expected_message = \
            'ResourceListFieldType cannot be used for request data'

        with self.assertRaisesMessage(NotImplementedError, expected_message):
            self.field_type.clean_value('abc')

    def test_str(self):
        """Testing ResourceListFieldType.__str__"""
        self.assertEqual(str(self.field_type),
                         'List of UserResource')


class StringFieldTypeTests(TestCase):
    """Unit tests for djblets.webapi.fields.StringFieldType."""

    def setUp(self):
        super(StringFieldTypeTests, self).setUp()

        self.field_type = StringFieldType({})

    def test_get_value_from_data(self):
        """Testing StringFieldType.get_value_from_data"""
        value = self.field_type.get_value_from_data(
            name='var',
            fields_data={
                'var': 'abc',
            },
            files_data={
                'var': 'def',
            })
        self.assertEqual(value, 'abc')

    def test_clean_value_with_valid_values(self):
        """Testing StringFieldType.clean_value with valid values"""
        self.assertEqual(self.field_type.clean_value('abc'), 'abc')
        self.assertEqual(self.field_type.clean_value(True), 'True')
        self.assertEqual(self.field_type.clean_value(100), '100')

    def test_str(self):
        """Testing StringFieldType.__str__"""
        self.assertEqual(str(self.field_type), 'String')



"""A specialization of pipeline's ES6Compiler."""

from pipeline.compilers.es6 import ES6Compiler as PipelineES6Compiler


class ES6Compiler(PipelineES6Compiler):
    """An ES6 Compiler which compiles .es6.js to .js.

    To use this, add the path to this class to
    ``settings.PIPELINE['COMPILERS']``.
    """

    input_extension = 'es6.js'

    def match_file(self, path):
        """Return whether the given path should use this compiler.

        Args:
            path (unicode):
                The source path.

        Returns:
            bool:
            Whether this compiler should be used for the given path.
        """
        return path.endswith(self.input_extension)

    def output_path(self, path, extension):
        """Return the path of the output file for this compiler.

        Args:
            path (unicode):
                The input path.

            extension (unicode):
                The target file extension.

        Returns:
            unicode:
            The input path with .es6.js renamed to .js.
        """
        assert path.endswith(self.input_extension)
        return path[:-len(self.input_extension)] + extension



"""URLs for extensions (used in the admin UI)."""

from django.urls import path

from djblets.extensions.views import extension_list


urlpatterns = [
    path('', extension_list, name='extension-list'),
]



"""Middleware for managing site configurations."""

from __future__ import annotations

from django.http import HttpRequest
from django.utils.deprecation import MiddlewareMixin

from djblets.siteconfig.models import SiteConfiguration


class SettingsMiddleware(MiddlewareMixin):
    """Middleware for performing expiration checks for site configuration.

    This will check the site configuration before each request is handled in
    order to see if it has expired, ensuring that the the request is able to
    work with the most up-to-date settings from the database.
    """

    def process_request(
        self,
        request: HttpRequest,
    ) -> None:
        """Process the HTTP request.

        This will perform an expiration check for the site configurations.

        Args:
            request (django.http.HttpRequest):
                The HTTP request being processed.
        """
        SiteConfiguration.objects.check_expired()



"""Utility functions for working with template contexts."""

from django.core.exceptions import ImproperlyConfigured
from django.template import engines


def get_default_template_context_processors(engine_name='django'):
    """Return all default template context processors for an engine.

    Args:
        engine_name (unicode):
            This specifies which template engine's context processors to
            return.

    Returns:
        list of callable:
        The list of template context processors.

    Raises:
        django.core.exceptions.ImproperlyConfigured:
            The specified template engine wasn't valid on this version of
            Django.
    """
    try:
        template_engine = engines[engine_name]
    except KeyError:
        raise ImproperlyConfigured(
            'The "%s" template engine must be defined.'
            % engine_name)

    return template_engine.engine.template_context_processors



from django.core.management.base import BaseCommand, CommandError
from django.utils.translation import gettext as _

from djblets.siteconfig.models import SiteConfiguration


class Command(BaseCommand):
    """Displays a setting in the site configuration."""

    def add_arguments(self, parser):
        """Add arguments to the command.

        Args:
            parser (object):
                The argument parser to add to.
        """
        parser.add_argument(
            '--key',
            action='store',
            dest='key',
            help=_('The existing key to display (dot-separated)'))

    def handle(self, *args, **options):
        siteconfig = SiteConfiguration.objects.get_current()

        key = options['key']

        if key is None:
            raise CommandError(_('--key must be provided'))

        path = key.split('.')
        node = siteconfig.settings
        valid_key = True
        key_basename = ''

        for item in path[:-1]:
            try:
                node = node[item]
            except KeyError:
                valid_key = False

        if valid_key:
            key_basename = path[-1]

            if key_basename not in node:
                valid_key = False

        if not valid_key:
            raise CommandError(_("'%s' is not a valid settings key") % key)

        value = node[key_basename]

        # None and boolean values are printed the same way that the JSON
        # serialization in list-siteconfig prints them, rather than the way
        # Python prints them.
        if value is None:
            value = 'null'
        elif isinstance(value, bool):
            if value:
                value = 'true'
            else:
                value = 'false'

        self.stdout.write('%s' % value)



from importlib import import_module

from django.db import connection

from djblets.testing.testcases import TestCase


# We need to ensure that the following module is imported so that the debug
# wrapper is replaced with our own.
import_module('djblets.log.middleware')


class CursorDebugWrapperTests(TestCase):
    """Unit tests for djblets.log.middleware.CursorDebugWrapper."""

    def test_execute(self):
        """Testing CursorDebugWrapper.execute"""
        with self.assertNumQueries(1):
            connection.cursor().execute(
                'INSERT INTO django_site (name, domain) VALUES (%s, %s)',
                ('site1', 'domain1.com'))

        self.assertIn('stack', connection.queries_log[-1])

    def test_executemany(self):
        """Testing CursorDebugWrapper.executemany"""
        with self.assertNumQueries(1):
            connection.cursor().executemany(
                'INSERT INTO django_site (name, domain) VALUES (%s, %s)',
                [
                    ('site1', 'domain1.com'),
                    ('site2', 'domain2.com'),
                ])

        self.assertIn('stack', connection.queries_log[-1])



"""Basic utility functions for authentication.

This contains some validation functions that may be useful for forms.
"""

from django.utils.translation import gettext as _


def validate_test_cookie(form, request):
    """Validate that the test cookie was properly set in a prior request.

    If the test cookie was not set, the given form's ``submit`` field will
    be set with an error saying that cookies must be enabled.

    Args:
        form (django.forms.Form):
            The form using the validator.

        request (django.http.HttpRequest):
            The HTTP request containing the test cookie.
    """
    if not request.session.test_cookie_worked():
        form.errors['submit'] = form.error_class(
            [_('Cookies must be enabled.')])


def validate_old_password(form, user, field_name='password'):
    """Validate that the password given on a form was valid.

    This is intended for Password Change forms, and will check that the
    specified password in the form matches the user's current password.

    Args:
        form (django.forms.Form):
            The form using the validator and containing the field.

        user (django.contrib.auth.models.User):
            The user whose password is being changed.

        field_name (unicode, optional):
            The name of the password field.
    """
    if (not form.errors.get(field_name) and
        not user.check_password(form.data.get(field_name))):
        form.errors[field_name] = form.error_class([_('Incorrect password.')])



"""Unit tests for CheckRequestMethodViewMixin."""

from django.http import HttpResponse
from django.test.client import RequestFactory
from django.views.generic.base import View

from djblets.testing.testcases import TestCase
from djblets.views.generic.base import CheckRequestMethodViewMixin


class BaseTestView(View):
    def dispatch(self, request, *args, **kwargs):
        # Note that we're not chaining up to View.dispatch, so that we don't
        # invoke its own method checks.
        return getattr(self, request.method.lower())(request, *args, **kwargs)


class CheckRequestMethodViewMixinTests(TestCase):
    """Unit tests for CheckRequestMethodViewMixin."""

    def test_dispatch_with_allowed(self):
        """Testing CheckRequestMethodViewMixin.dispatch with HTTP method
        allowed
        """
        class MyView(CheckRequestMethodViewMixin, BaseTestView):
            def get(self, *args, **kwargs):
                return HttpResponse('ok')

        view = MyView.as_view()
        response = view(RequestFactory().request())

        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.content, b'ok')

    def test_dispatch_with_not_allowed(self):
        """Testing CheckRequestMethodViewMixin.dispatch with HTTP method
        not allowed
        """
        class MyView(CheckRequestMethodViewMixin, BaseTestView):
            def get(self, *args, **kwargs):
                return HttpResponse('ok')

        view = MyView.as_view()
        response = view(RequestFactory().request(REQUEST_METHOD='POST'))

        self.assertEqual(response.status_code, 405)
        self.assertNotEqual(response.content, b'ok')



"""Utilities for generating resource trees for testing."""

from collections import namedtuple

from djblets.extensions.resources import (
    ExtensionResource as BaseExtensionResource)
from djblets.webapi.resources import WebAPIResource
from djblets.webapi.resources.root import RootResource as BaseRootResource


ResourceTree = namedtuple('ResourceTree',
                          ('base_resource', 'root_resource', 'resources'))


def make_resource_tree(mixins=None, extension_manager=None,
                       allow_anonymous_access=True):
    """Create and return a generated resource tree.

    Args:
        mixins (list of type):
            A list of mixin classes for the resources.

        extension_manager (djblets.extensions.manager.ExtensionManager,
                           optional):
            An optional extension manager. If this is provided, the generated
            tree will have an extension resource.

        allow_anonymous_access (bool, optional):
            Whether or not anonymous access should be allowed.

    Returns:
        ResourceTree:
        The generated resource tree.
    """
    if mixins is None:
        mixins = []

    bases = tuple(mixins + [WebAPIResource])

    def has_access_permissions(self, request, *args, **kwargs):
        return allow_anonymous_access or request.user.is_authenticated

    base_resource = type('BaseResource', bases, {
        'has_access_permissions': has_access_permissions,
    })

    resources = []

    if extension_manager:
        class ExtensionResource(base_resource, BaseExtensionResource):
            pass

        resources.append(ExtensionResource(extension_manager))

    class ItemChildResource(base_resource):
        """An item child resource."""

        name = 'item-child'
        allowed_methods = ('GET', 'POST', 'PUT')

    class ForbiddenResource(base_resource):
        """A resource that cannot be accessed via OAuth2 or API tokens."""

        name = 'forbidden'
        singleton = True
        allowed_methods = ('GET', 'POST', 'PUT', 'DELETE')
        oauth2_token_access_allowed = False
        api_token_access_allowed = False

    class ListChildResource(base_resource):
        """A list child resource."""

        name = 'list-child'
        allowed_methods = ('GET', 'DELETE')

    class ParentResource(base_resource):
        """A resource with two children."""

        name = 'parent'
        allowed_methods = ('GET',)

        item_child_resources = [ItemChildResource()]
        list_child_resources = [ListChildResource()]

    resources.extend([
        ParentResource(),
        ForbiddenResource(),
    ])

    class RootResource(base_resource, BaseRootResource):
        """The root of the resource tree."""

        def __init__(self, *args, **kwargs):
            super(RootResource, self).__init__(resources, *args, **kwargs)

    return ResourceTree(
        base_resource=base_resource,
        root_resource=RootResource(resources),
        resources=resources,
    )



"""Unit tests for djblets.util.templatetags.djblets_forms."""

import os

from django.conf import settings
from django.template import Context, Template
from pipeline.conf import settings as pipeline_settings

from djblets.testing.testcases import TestCase


class CompressedTagsTests(TestCase):
    """Unit tests for the {% compressed_* %} template tags."""

    def test_compressed_css_tag(self):
        """Testing {% compressed_css %}"""
        self._touch_files(['test.css', 'test.d41d8cd98f00.css'])

        pipeline_settings.STYLESHEETS = {
            'test': {
                'source_filenames': [],
                'output_filename': 'test.css',
            }
        }

        t = Template('{% load compressed %}'
                     '{% compressed_css "test" %}')

        self.assertHTMLEqual(
            t.render(Context({'test': 'test'})),
            '<link href="/test.d41d8cd98f00.css" media="all" rel="stylesheet"'
            ' type="text/css">')

    def test_compressed_js_tag(self):
        """Testing {% compressed_js %}"""
        self._touch_files(['test.js', 'test.d41d8cd98f00.js'])

        pipeline_settings.JAVASCRIPT = {
            'test': {
                'source_filenames': [],
                'output_filename': 'test.js',
            }
        }

        t = Template('{% load compressed %}'
                     '{% compressed_js "test" %}')

        self.assertHTMLEqual(
            t.render(Context({'test': 'test'})),
            '<script type="text/javascript" src="/test.d41d8cd98f00.js"'
            ' charset="utf-8"></script>')

    def _touch_files(self, filenames):
        """Create one or more empty static media files.

        Args:
            filenames (list of unicode):
                The list of static media files to create.
        """
        for filename in filenames:
            with open(os.path.join(settings.STATIC_ROOT, filename), 'w'):
                pass



"""Context processors for serial numbers used for front-end caching.

These context processors can be used to inject the :setting:`MEDIA_SERIAL`
and :setting:`AJAX_SERIAL` settings into templates so that links can be
built that take these caching serials into account.

See :py:mod:`djblets.cache.serials` for more information on building these
serials.
"""

from django.conf import settings


def media_serial(request):
    """Add ``MEDIA_SERIAL`` to template contexts.

    Exposes a media serial number that can be appended to a media filename
    in order to make a URL that can be cached forever without fear of change.
    The next time the file is updated and the server is restarted, a new
    path will be accessed and cached.

    This returns the value of :setting:`MEDIA_SERIAL`, which must either be
    set manually or ideally should be set to the value of
    :py:func:`djblets.cache.serials.generate_media_serial`.
    """
    return {'MEDIA_SERIAL': getattr(settings, "MEDIA_SERIAL", "")}


def ajax_serial(request):
    """Add ``AJAX_SERIAL`` to template contexts.

    Exposes a serial number that can be appended to filenames involving
    dynamic loads of URLs in order to make a URL that can be cached forever
    without fear of change.

    This returns the value of :setting:`AJAX_SERIAL`, which must either be
    set manually or ideally should be set to the value of
    :py:func:`djblets.cache.serials.generate_ajax_serial`.
    """
    return {'AJAX_SERIAL': getattr(settings, "AJAX_SERIAL", "")}



"""
Sphinx plugins for special links in the Release Notes.
"""
from docutils import nodes, utils


def setup(app):
    app.add_config_value('bugtracker_url', '', True)
    app.add_role('bug', bug_role)
    app.add_role('cve', cve_role)


def bug_role(role, rawtext, text, linenum, inliner, options={}, content=[]):
    try:
        bugnum = int(text)
        if bugnum <= 0:
            raise ValueError
    except ValueError:
        msg = inliner.reporter.error(
            'Bug number must be a number greater than or equal to 1; '
            '"%s" is invalid.' % text,
            line=linenum)
        prb = inliner.problematic(rawtext, rawtext, msg)
        return [prb], [msg]

    bugtracker_url = inliner.document.settings.env.config.bugtracker_url

    if not bugtracker_url or '%s' not in bugtracker_url:
        msg = inliner.reporter.error('bugtracker_url must be configured.',
                                     line=linenum)
        prb = inliner.problematic(rawtext, rawtext, msg)
        return [prb], [msg]

    ref = bugtracker_url % bugnum
    node = nodes.reference(rawtext, 'Bug #' + utils.unescape(text),
                           refuri=ref, **options)

    return [node], []


def cve_role(role, rawtext, text, linenum, inliner, options={}, content=[]):
    ref = 'http://cve.mitre.org/cgi-bin/cvename.cgi?name=CVE-%s' % text
    node = nodes.reference(rawtext, 'CVE-' + utils.unescape(text),
                           refuri=ref, **options)

    return [node], []



"""Encryption/decryption utilities.

Version Added:
    3.0
"""

import base64
import os
from typing import AnyStr, Iterable, Iterator, Optional, Union, cast

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
from django.conf import settings


AES_BLOCK_SIZE = cast(int, algorithms.AES.block_size) // 8


def _create_cipher(
    iv: bytes,
    *,
    key: Optional[bytes] = None,
) -> Cipher:
    """Create a cipher for use in symmetric encryption/decryption.

    This will use AES encryption in CFB mode (using an 8-bit shift register)
    and a random IV.

    Version Added:
        3.0

    Args:
        iv (bytes):
            The random IV to use for the cipher.

        key (bytes, optional):
            The encryption key to use.

    Returns:
        cryptography.hazmat.primitives.cipher.Cipher:
        The cipher to use for encryption/decryption.

    Raises:
        ValueError:
            The encryption key was not in the right format.
    """
    if key is None:
        key = get_default_aes_encryption_key()

    if not isinstance(key, bytes):
        raise TypeError('The encryption key must be of type "bytes", not "%s"'
                        % type(key))

    return Cipher(algorithms.AES(key),
                  modes.CFB8(iv),
                  default_backend())


def get_default_aes_encryption_key() -> bytes:
    """Return the default AES encryption key for the install.

    The default key is the first 16 characters (128 bits) of
    :django:setting:`SECRET_KEY`.

    Version Added:
        3.0

    Returns:
        bytes:
        The default encryption key.
    """
    return settings.SECRET_KEY[:16].encode('utf-8')


def aes_encrypt(
    data: Union[bytes, str],
    *,
    key: Optional[bytes] = None,
) -> bytes:
    """Encrypt data using AES encryption.

    This uses AES encryption in CFB mode (using an 8-bit shift register) and a
    random IV (which will be prepended to the encrypted value). The encrypted
    data will be decryptable using the :py:func:`aes_decrypt` function.

    Version Added:
        3.0

    Args:
        data (bytes or str):
            The data to encrypt. If a Unicode string is passed in, it will be
            encoded to UTF-8 first.

        key (bytes, optional):
            The optional custom encryption key to use. If not supplied, the
            default encryption key (from
            :py:func:`get_default_aes_encryption_key)` will be used.

    Returns:
        bytes:
        The resulting encrypted value, with the random IV prepended.

    Raises:
        ValueError:
            The encryption key was not in the right format.
    """
    if isinstance(data, str):
        data = data.encode('utf-8')

    iv = os.urandom(AES_BLOCK_SIZE)
    cipher = _create_cipher(iv, key=key)
    encryptor = cipher.encryptor()

    return iv + encryptor.update(data) + encryptor.finalize()


def aes_encrypt_base64(
    data: AnyStr,
    *,
    key: Optional[bytes] = None,
) -> str:
    """Encrypt data and encode as Base64.

    The result will be encrypted using AES encryption in CFB mode (using an
    8-bit shift register), and serialized into Base64.

    Version Added:
        3.0

    Args:
        data (bytes or str):
            The data to encrypt. If a Unicode string is passed in, it will
            be encoded to UTF-8 first.

        key (bytes, optional):
            The optional custom encryption key to use. If not supplied, the
            default encryption key (from
            :py:func:`get_default_aes_encryption_key)` will be used.

    Returns:
        str:
        The encrypted password encoded in Base64.

    Raises:
        ValueError:
            The encryption key was not in the right format.
    """
    return base64.b64encode(aes_encrypt(data, key=key)).decode('utf-8')


def aes_encrypt_iter(
    data_iter: Iterable[Union[bytes, str]],
    *,
    key: Optional[bytes] = None,
) -> Iterator[bytes]:
    """Encrypt and yield data iteratively.

    This iterates through an iterable (a generator, list, or similar),
    yielding AES-encrypted batches of data. This can be used when streaming
    a source and yielding encrypted data to a file, HTTP response, across
    multiple cache keys, etc.

    The result can be decrypted either by joining together all the results
    or by passing the results to :py:func:`aes_decrypt_iter`.

    Args:
        data_iter (iterable):
            An iterator that yields byte strings or Unicode strings.

        key (bytes, optional):
            The optional custom encryption key to use. If not supplied, the
            default encryption key (from
            :py:func:`get_default_aes_encryption_key)` will be used.

    Yields:
        bytes:
        An encrypted block of data.

    Raises:
        ValueError:
            The encryption key was not in the right format.
    """
    iv = os.urandom(AES_BLOCK_SIZE)
    cipher = _create_cipher(iv, key=key)
    encryptor = cipher.encryptor()

    # We want the very first value to contain the iv, and the very last to
    # contain the finalizer. So we need to operate one ahead of the item from
    # the stream. We'll iterate through, grab an item, then yield the previous
    # one (prepending the iv if yielding the very first).
    prev_item = None

    for item in data_iter:
        if isinstance(item, str):
            item = item.encode('utf-8')

        encrypted_item = encryptor.update(item)

        if prev_item is not None:
            yield prev_item

            prev_item = encrypted_item
        else:
            prev_item = iv + encrypted_item

    # We can now follow up with the finalizer.
    if prev_item is not None:
        yield prev_item + encryptor.finalize()
    else:
        # We had absolutely nothing to yield. Let's just yield an empty
        # encrypted block.
        yield iv + encryptor.finalize()


def aes_decrypt(
    encrypted_data: bytes,
    *,
    key: Optional[bytes] = None,
) -> bytes:
    """Decrypt AES-encrypted data.

    This will decrypt an AES-encrypted value in CFB mode (using an 8-bit
    shift register). It expects the 16-byte cipher IV to be prepended to the
    string.

    This is intended as a counterpart for :py:func:`aes_encrypt`.

    Version Added:
        3.0

    Args:
        encrypted_data (bytes):
            The data to decrypt.

        key (bytes, optional):
            The optional custom encryption key to use. This must match the key
            used for encryption. If not supplied, the default encryption key
            (from :py:func:`get_default_aes_encryption_key)` will be used.

    Returns:
        bytes:
        The decrypted value.

    Raises:
        TypeError:
            One or more arguments had an invalid type.

        ValueError:
            The encryption key was not in the right format.
    """
    if not isinstance(encrypted_data, bytes):
        raise TypeError('The data to decrypt must be of type "bytes", not "%s"'
                        % (type(encrypted_data)))

    cipher = _create_cipher(encrypted_data[:AES_BLOCK_SIZE],
                            key=key)
    decryptor = cipher.decryptor()

    return (decryptor.update(encrypted_data[AES_BLOCK_SIZE:]) +
            decryptor.finalize())


def aes_decrypt_base64(
    encrypted_data: AnyStr,
    *,
    key: Optional[bytes] = None,
) -> str:
    """Decrypt an encrypted value encoded in Base64.

    This will decrypt a Base64-encoded encrypted value (from
    :py:func:`aes_encrypt_base64`) into a string.

    Version Added:
        3.0

    Args:
        encrypted_data (bytes or str):
            The Base64-encoded encrypted data to decrypt.

        key (bytes, optional):
            The optional custom encryption key to use. This must match the key
            used for encryption. If not supplied, the default encryption key
            (from :py:func:`get_default_aes_encryption_key)` will be used.

    Returns:
        str:
        The resulting decrypted data.

    Raises:
        ValueError:
            The encryption key was not in the right format.
    """
    return (
        aes_decrypt(base64.b64decode(encrypted_data),
                    key=key)
        .decode('utf-8')
    )


def aes_decrypt_iter(
    encrypted_iter: Iterable[bytes],
    *,
    key: Optional[bytes] = None,
) -> Iterator[bytes]:
    """Decrypt and yield data iteratively.

    This iterates through an iterable (a generator, list, or similar),
    decrypting items and yielding the decrypted values. This can be used when
    streaming an encrypted source and yielding the decrypted results to a file,
    HTTP response, across multiple cache keys, etc.

    Args:
        encrypted_iter (iterable):
            An iterator that yields AES-encrypted data as byte strings.

        key (bytes, optional):
            The optional custom encryption key to use. If not supplied, the
            default encryption key (from
            :py:func:`get_default_aes_encryption_key)` will be used.

    Yields:
        bytes:
        A decrypted block of data.

    Raises:
        ValueError:
            The encryption key was not in the right format or the encrypted
            data was invalid.
    """
    # Ensure we're working with an actual iterator now, not just something
    # iterable. We need to ensure we're not starting iteration over when we
    # loop through a second time.
    encrypted_iter = iter(encrypted_iter)

    # We need to read enough to get the IV (the Initialization Vector at the
    # start of the encrypted data). We'll keep yielding items until we find
    # it, or until we have nothing left to read.
    #
    # If we don't receive enough for the IV, cryptography will raise a
    # ValueError below (same as when passing an incomplete payload to
    # aes_decrypt()).
    iv_buf = b''

    for item in encrypted_iter:
        iv_buf += item

        if len(iv_buf) >= AES_BLOCK_SIZE:
            break

    # Create the cipher as normal.
    cipher = _create_cipher(iv_buf[:AES_BLOCK_SIZE],
                            key=key)
    decryptor = cipher.decryptor()

    # Start with the data after the IV, and then go through the iterator.
    # Like with aes_encrypt_iter(), we're going to fetch the next item and
    # *then* yield the previous item. The reason is that we want to ensure
    # the finalizer is part of the last item yielded.
    prev_item = iv_buf[AES_BLOCK_SIZE:] or None

    for item in encrypted_iter:
        if prev_item is not None:
            yield decryptor.update(prev_item)

        prev_item = item

    if prev_item is not None:
        yield decryptor.update(prev_item) + decryptor.finalize()
    else:
        yield decryptor.finalize()



"""Base class for token generators.

Version Added:
    3.0
"""

from typing import Any, Dict, Optional


class BaseTokenGenerator:
    """A base class for token generators.

    These are used to generate and validate access tokens.

    At the very least, subclasses must set:

    * :py:attr:`token_generator_id`

    as well as implement the :py:meth:`create_token` and
    :py:meth:`validate_token` methods.

    Version Added:
        3.0
    """

    #: The unique ID of the token generator, for registration purposes.
    #:
    #: Type:
    #:     str
    token_generator_id: Optional[str] = None

    def create_token(
        self,
        token_info: Dict[str, Any] = {},
        **kwargs,
    ) -> str:
        """Create a new token.

        Subclasses must override this to actually create the token.

        Args:
            token_info (dict, optional):
                A dictionary that contains information about the
                token.

            **kwargs (dict):
                Additional keyword arguments.

        Returns:
            str:
            The token that was created.

        Raises:
            NotImplementedError:
                The subclass did not implement this method.
        """
        raise NotImplementedError

    def validate_token(
        self,
        token: str,
        token_info: Dict[str, Any] = {},
        **kwargs,
    ) -> bool:
        """Validate the token.

        This should return a bool indicating whether the given token
        is a valid token that could have been generated by this generator.

        Subclasses must override this to actually validate the token.

        Args:
            token (str):
                The token to validate.

            token_info (dict, optional):
                A dictionary that contains information about the
                token.

            **kwargs (dict):
                Additional keyword arguments.

        Returns:
            bool:
            Whether the token is a valid token from this generator.

        Raises:
            NotImplementedError:
                The subclass did not implement this method.
        """
        raise NotImplementedError



"""Image-related template tags."""

import io
import logging
import os
import re

from django import template
from django.template import TemplateSyntaxError
from django.utils.html import format_html, format_html_join
from django.utils.translation import gettext as _
from PIL import Image
from PIL.Image import registered_extensions

from djblets.util.decorators import blocktag


logger = logging.getLogger(__name__)
register = template.Library()


def save_image_to_storage(image, storage, filename):
    """Save an image to storage."""
    f = storage.open(filename, mode='w+b')
    image.save(f, 'png')
    f.close()


@register.simple_tag
def crop_image(f, x, y, width, height):
    """
    Crops an image at the specified coordinates and dimensions, returning the
    resulting URL of the cropped image.
    """
    filename = f.name
    storage = f.storage
    basename = filename

    if filename.find(".") != -1:
        basename = filename.rsplit('.', 1)[0]
    new_name = '%s_%d_%d_%d_%d.png' % (basename, x, y, width, height)

    if not storage.exists(new_name):
        try:
            f = storage.open(filename)
            data = io.BytesIO(f.read())
            f.close()

            image = Image.open(data)
            image = image.crop((x, y, x + width, y + height))

            save_image_to_storage(image, storage, new_name)
        except (IOError, KeyError) as e:
            logger.exception('Error cropping image file %s at %d, %d, %d, %d '
                             'and saving as %s: %s',
                             filename, x, y, width, height, new_name, e)
            return ""

    return storage.url(new_name)


@register.filter
def thumbnail(f, size='400x100'):
    """Create a thumbnail of the given image.

    This will create a thumbnail of the given ``file`` (a Django FileField or
    ImageField) with the given size. Size can either be a string of WxH (in
    pixels), or a 2-tuple. If the size is a tuple and the second part is None,
    it will be calculated to preserve the aspect ratio.

    If the image format is not registered in PIL the thumbnail is not generated
    and returned as-is.

    This will return the URL to the stored thumbnail.
    """
    storage = f.storage
    ext = os.path.splitext(f.name)[1].lower()
    if ext not in registered_extensions():
        return storage.url(f.name)

    if isinstance(size, str):
        x, y = (int(x) for x in size.split('x'))
        size_str = size
    elif isinstance(size, tuple):
        x, y = size

        if y is None:
            size_str = '%d' % x
        else:
            size_str = '%dx%d' % (x, y)
    else:
        raise ValueError('Thumbnail size "%r" could not be be parsed', size)

    filename = f.name
    if filename.find(".") != -1:
        basename, format = filename.rsplit('.', 1)
        miniature = '%s_%s.%s' % (basename, size_str, format)
    else:
        basename = filename
        miniature = '%s_%s' % (basename, size_str)

    if not storage.exists(miniature):
        try:
            f = storage.open(filename, 'rb')
            data = io.BytesIO(f.read())
            f.close()

            image = Image.open(data)

            if y is None:
                x = min(image.size[0], x)

                # Calculate height based on width
                y = int(x * (image.size[1] / image.size[0]))

            # Pillow is aggressively deprecating Image.ANTIALIAS. Conditionally
            # attempt to use the new name.
            if hasattr(Image, 'Resampling'):
                # 9.1.0+
                antialias = Image.Resampling.LANCZOS
            else:
                # 9.0.x and below
                antialias = Image.ANTIALIAS

            image.thumbnail((x, y), antialias)

            save_image_to_storage(image, storage, miniature)
        except (IOError, KeyError) as e:
            logger.exception('Error thumbnailing image file %s and saving '
                             'as %s: %s',
                             filename, miniature, e)
            return ""

    return storage.url(miniature)


def build_srcset(sources):
    """Return the source set attribute value for the given sources.

    The resulting sources will be sorted by value, with the pixel density
    (``x``) values coming before width (``w``) values.

    Args:
        sources (dict):
            A mapping of descriptors (e.g., ``'2x'`` or ``'512w'``) that
            describe the requirement for the source to be shown to URLs.

    Returns:
        unicode:
        The returned ``srcset`` attribute value.
    """
    sources_info = []

    for descriptor, url in sources.items():
        if not url:
            continue

        if not descriptor:
            descriptor = '1x'

        valid = descriptor.endswith(('x', 'w'))

        if valid:
            try:
                sources_info.append((descriptor, float(descriptor[:-1]), url))
            except ValueError:
                valid = False

        if not valid:
            raise ValueError(_('"%s" is not a valid srcset size descriptor.')
                             % descriptor)

    # Sort the sources such that 'x' descriptors are always before 'w'
    # descriptors, and in numerical order.
    descriptor_sort_values = {
        'x': -1,
        'w': 1,
    }

    sources_info = sorted(
        sources_info,
        key=lambda source: (descriptor_sort_values[source[0][-1]], source[1]))

    return format_html_join(
        ', ',
        '{0} {1}',
        (
            (url, descriptor)
            for descriptor, descriptor_value, url in sources_info
        ))


@register.simple_tag
def srcset(sources):
    """Render the source set attribute value for the given sources.

    The resulting sources will be sorted by value, with the pixel density
    (``x``) values coming before width (``w``) values.

    Args:
        sources (dict):
            A mapping of descriptors (e.g., ``'2x'`` or ``'512w'``) that
            describe the requirement for the source to be shown to URLs.

    Returns:
        unicode:
        The rendered ``srcset`` attribute value.
    """
    try:
        return build_srcset(sources)
    except ValueError as e:
        raise TemplateSyntaxError(str(e))


@register.tag
@blocktag(end_prefix='end_')
def image_source_attrs(context, nodelist, *options):
    """Render source attributes for an image tag.

    This will render ``src="..." srcset="..."`` attributes for an ``<img>``
    tag, based on the sources provided in the tag's content. There should be
    one source definition per line (with an optional trailing comma) in the
    form of::

        <descriptor> <URL>

    These will get turned into a ``srcset``, and the ``1x`` descriptor (which
    is required) will be set as the ``src`` attribute.

    Args:
        block_content (unicode):
            The block content containing image sources.

    Returns:
        Attributes for the ``<img>`` tag.

    Example:
        .. code-block:: html+django

           <img {% image_source_attrs %}
                1x {%  static "images/myimage.png" %}
                2x {%  static "images/myimage@2x.png" %}
                3x {%  static "images/myimage@3x.png" %}
                {% end_image_source_attrs %}>
    """
    content = nodelist.render(context).strip()

    try:
        sources = {}

        for source in re.split(r',|\n+', content):
            source = source.strip()

            if source:
                descriptor, url = source.split(' ', 1)
                sources[descriptor.strip()] = url.strip()

    except ValueError:
        raise TemplateSyntaxError(_(
            'The source definition passed to {% image_source_attrs %} is '
            'not structured correctly. Make sure that there is one source '
            'definition per line and that it contains a descriptor and a '
            'URL.'))

    try:
        src_value = sources['1x']
    except KeyError:
        raise TemplateSyntaxError(_(
            'The source definition passed to {% image_source_attr %} must '
            'contain a "1x" descriptor.'))

    return format_html('src="{0}" srcset="{1}"', src_value,
                       build_srcset(sources))



"""Extension testing support.

This provides handy imports for extension testing classes:

* :py:class:`djblets.extensions.testing.testcases.ExtensionTestCaseMixin`
"""

from djblets.extensions.testing.testcases import ExtensionTestCaseMixin


__all__ = [
    'ExtensionTestCaseMixin',
]

__autodoc_excludes__ = __all__



"""Base support for djblets.extensions unit tests."""

import pkg_resources

from djblets.extensions.manager import ExtensionManager


class FakeEntryPoint(object):
    """A fake entry point.

    This is modelled after :py:class:`pkg_resources.EntryPoint`.
    """

    def __init__(self, value, project_name, version='1.0', **metadata_kwargs):
        """Initialize the FakeEntryPoint.

        Args:
            value (object):
                The value to be returned when the entry point is loaded.

            project_name (unicode):
                The project name. This will be set in the metadata and as
                the distribution's name.

            **metadata_kwargs (dict):
                Keyword arguments to pass to the associated
                :py:class:`FakeProvider` constructor.
        """
        self._value = value
        self.dist = pkg_resources.Distribution(
            project_name=project_name,
            version=version,
            metadata=FakeProvider(project_name=project_name,
                                  version=version,
                                  **metadata_kwargs))

    def load(self):
        """Load the entry point.

        Returns:
            object: The value specified at initialization time.
        """
        return self._value


class FakeProvider(pkg_resources.DefaultProvider):
    """A fake provider for a distribution."""

    egg_info = '/fake/path'

    def __init__(self,
                 metadata=None,
                 author='Example Author',
                 author_email='author@example.com',
                 description='Test description\u2049',
                 home_page='http://example.com',
                 project_name='ExampleProject',
                 license_name='Drivers',
                 summary='Test summary',
                 version='1.0'):
        """Initialize the FakeDistribution.

        Args:
            metadata (dict, optional):
                Metadata to assign to the provider. If this is provided, it
                overrides all other arguments.

            author (unicode, optional):
                The package author.

            author_email (unicode, optional):
                The package author's e-mail address.

            description (unicode, optional):
                The package description.

            home_page (unicode, optional):
                The package's URL.

            project_name (unicode, optional):
                The package's name.

            license_name (unicode, optional):
                The name of the package license.

            summary (unicode, optional):
                The package summary.
        """
        if metadata is not None:
            self.metadata = metadata
        else:
            self.metadata = {
                'Author': author,
                'Author-email': author_email,
                'Description': description,
                'Home-page': home_page,
                'Name': project_name,
                'License': license_name,
                'Summary': summary,
                'Version': version,
            }

    def _get(self, path):
        """Return the metadata content.

        This is the method that package resource providers must override to
        return metadata content for the package. It's expected to return
        byte strings, which will then be handled through the normal metadata
        functions.

        Returns:
            bytes:
            The package metadata content.
        """
        return ''.join(
            '%s: %s\n' % (field_name, value)
            for field_name, value in self.metadata.items()
        ).encode('utf-8')


class MyTestExtensionManager(ExtensionManager):
    """An extension manager for testing.

    The entry points are provided at initialization time.
    """

    def __init__(self, entry_points=None, *args, **kwargs):
        """Initialize the MyTestExtensionManager.

        Args:
            entry_points (list):
                A list of :py:class:`FakeEntryPoint`s.

            *args (tuple):
                Additional positional arguments to pass to the base class'
                constructor.

            **kwargs (dict):
                Additional keyword arguments to pass to the base class'
                constructor.
        """
        super(MyTestExtensionManager, self).__init__(*args, **kwargs)
        self._entry_points = entry_points or []

    def _entrypoint_iterator(self):
        """Return the entry points.

        Returns:
            list: The entry points.
        """
        return self._entry_points


class ExtensionTestsMixin(object):
    """Mixin for Djblets extension-related unit tests.

    This is used to help set up an extension and extension manager to test
    with.

    This shouldn't generally be used by third-parties writing unit tests for
    their extensions. For those, see
    :py:mod:`djblets.extensions.testing.testcases`.
    """

    #: The project name to use in the entrypoint.
    test_project_name = 'TestProject'

    def setUp(self):
        """Set up state before a test run."""
        super(ExtensionTestsMixin, self).setUp()

        self.manager = None

    def tearDown(self):
        """Tear down state after a test run.

        This will properly ensure that the extension manager, if one was
        created, will clear all state and shut down.
        """
        super(ExtensionTestsMixin, self).tearDown()

        if self.manager:
            self.manager.clear_sync_cache()
            self.manager.shutdown()

    def setup_extension(self, extension_cls, enable=True, manager_key='tests'):
        """Set up an extension for use in a test.

        This will register the class in a new extension manager and then
        enable or otherwise instantiate the extension, returning the instance.

        Args:
            extension_cls (type):
                The extension class to register.

            enable (bool, optional):
                Whether the returned extension should be enabled.

            manager_key (unicode, optional):
                The key to use for the extension manager.

        Returns:
            djblets.extensions.extension.Extension:
            The resulting extension instance.
        """
        fake_entry_point = FakeEntryPoint(extension_cls,
                                          project_name=self.test_project_name)

        self.manager = MyTestExtensionManager([fake_entry_point], manager_key)
        self.manager.load()

        if enable:
            return self.manager.enable_extension(extension_cls.id)
        else:
            return extension_cls(self.manager)



"""Unit tests for extension middleware."""

import kgb
from django.test import override_settings

from djblets.extensions.extension import Extension
from djblets.extensions.testing import ExtensionTestCaseMixin
from djblets.testing.testcases import TestCase


MIDDLEWARE = [
    'django.middleware.common.CommonMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'djblets.extensions.middleware.ExtensionsMiddleware',
    'djblets.extensions.middleware.ExtensionsMiddlewareRunner',
]


class Middleware(object):
    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        return self.get_response(request)


class MyTestExtension(Extension):
    middleware = [
        'djblets.extensions.tests.test_middleware.Middleware',
    ]


class MiddlewareTests(kgb.SpyAgency, ExtensionTestCaseMixin, TestCase):
    """Unit tests for extension middleware."""

    @override_settings(MIDDLEWARE=MIDDLEWARE)
    def test_extension_middleware(self):
        """Testing extension middleware"""
        self.spy_on(Middleware.__call__, owner=Middleware)
        self.setup_extension(MyTestExtension)
        self.client.get('/')
        self.assertSpyCalled(Middleware.__call__)



from django.conf import settings
from django.core.cache import cache

from djblets.integrations.manager import shutdown_integration_managers
from djblets.testing.testcases import TestCase, TestModelsLoaderMixin


class IntegrationsTestCase(TestModelsLoaderMixin, TestCase):
    """Base class for unit tests that work with integrations."""

    tests_app = 'djblets.integrations.tests'

    def setUp(self):
        super(IntegrationsTestCase, self).setUp()

        self.old_middleware_classes = list(settings.MIDDLEWARE)
        settings.MIDDLEWARE = self.old_middleware_classes + [
            'djblets.integrations.middleware.IntegrationsMiddleware',
        ]

        cache.clear()

    def tearDown(self):
        super(IntegrationsTestCase, self).tearDown()

        settings.MIDDLEWARE = self.old_middleware_classes

        shutdown_integration_managers()



"""Unit tests for the web API token auth backend."""

import datetime
import logging
import time

import kgb
from django.contrib.auth.models import User
from django.contrib.sessions.middleware import SessionMiddleware
from django.core.cache import cache
from django.http import HttpResponse
from django.test.client import RequestFactory
from django.test.utils import override_settings
from django.utils import timezone

from djblets.auth.ratelimit import _get_time_int
from djblets.testing.testcases import TestCase, TestModelsLoaderMixin
from djblets.webapi.auth.backends.api_tokens import (TokenAuthBackendMixin,
                                                     WebAPITokenAuthBackend)
from djblets.webapi.models import BaseWebAPIToken


class MyTestWebAPITokenModel(BaseWebAPIToken):
    """Mock WebAPI Token Model for testing purposes."""


class MyTestTokenAuthBackend(TokenAuthBackendMixin):
    """Mock Token Auth Backend for testing purposes."""

    api_token_model = MyTestWebAPITokenModel


@override_settings(AUTHENTICATION_BACKENDS=(
    'djblets.webapi.tests.test_api_auth_backend.MyTestTokenAuthBackend',
))

class WebAPITokenAuthBackendTests(kgb.SpyAgency,
                                  TestModelsLoaderMixin,
                                  TestCase):
    """Unit tests for the WebAPITokenAuthBackend."""

    tests_app = 'djblets.webapi.tests'

    def setUp(self):
        super(WebAPITokenAuthBackendTests, self).setUp()
        self.api_token_auth_backend = WebAPITokenAuthBackend()
        self.request = RequestFactory().get('/')

        middleware = SessionMiddleware(lambda request: HttpResponse(''))
        middleware(self.request)

    def tearDown(self):
        super(WebAPITokenAuthBackendTests, self).tearDown()
        cache.clear()

    def test_get_credentials_no_token_realm(self):
        """Testing Token Auth get_credentials with non-token realm"""
        self.request.META['HTTP_AUTHORIZATION'] = 'NotTokenRealm tokenHere'
        result = self.api_token_auth_backend.get_credentials(self.request)
        self.assertIsNone(result)

    def test_get_credentials_missing_credentials(self):
        """Testing Token Auth get_credentials with missing token"""
        self.request.META['HTTP_AUTHORIZATION'] = 'token'
        logger = logging.getLogger('djblets.webapi.auth.backends.api_tokens')
        self.spy_on(logger.warning)

        result = self.api_token_auth_backend.get_credentials(self.request)
        self.assertIsNone(result)

        warning_message = logger.warning.spy.last_call.args[0]
        self.assertTrue(warning_message.startswith(
            'WebAPITokenAuthBackend: Missing token in HTTP_AUTHORIZATION '
            'header'))

    def test_get_credentials_valid_credentials(self):
        """Testing Token Auth get_credentials with valid credentials"""
        token = 'myToken'
        self.request.META['HTTP_AUTHORIZATION'] = 'token %s' % token

        result = self.api_token_auth_backend.get_credentials(self.request)
        self.assertEqual(result, {'token': token})

    def test_authenticate_valid_credentials(self):
        """Testing Token Auth authenticate succeeds with valid credentials"""
        self.spy_on(timezone.now, op=kgb.SpyOpReturn(timezone.now()))

        token = 'invalidtoken123'
        self.user = User.objects.create_user(username='testuser')
        token_object = MyTestWebAPITokenModel.objects.create(user=self.user,
                                                             token=token)
        self.request.user = User()
        self.request.META['HTTP_AUTHORIZATION'] = 'token %s' % token

        result = self.api_token_auth_backend.authenticate(self.request)
        token_object.refresh_from_db(fields=('last_used',))

        self.assertEqual(result, (True, None, None))
        self.assertNotEqual(self.request.META['CSRF_COOKIE'], '')
        self.assertEqual(token_object.last_used, timezone.now())

    def test_authenticate_wrong_token(self):
        """Testing Token Auth authenticate failed with wrong token"""
        token = 'invalidtoken123'
        self.user = User.objects.create_user(username='testuser')
        MyTestWebAPITokenModel.objects.create(user=self.user, token=token)
        self.request.user = User()
        self.request.META['HTTP_AUTHORIZATION'] = 'token bad_token'

        result = self.api_token_auth_backend.authenticate(self.request)
        self.assertEqual(result, (False, None, None))

    @override_settings(LOGIN_LIMIT_RATE='5/m')
    def test_authenticate_rate_limit(self):
        """Testing Token Auth authenticate failed and exceeding
        number of failed login attempts allowed with default rate limit
        """
        token = 'myToken'
        self.user = User.objects.create_user(username='testratelimit')
        MyTestWebAPITokenModel.objects.create(user=self.user, token=token)
        self.request.user = self.user

        # Send invalid token to count number of failed login attempts
        self.request.META['HTTP_AUTHORIZATION'] = 'token bad_token'

        # Make sure that the time does not change during this test.
        self.spy_on(_get_time_int, op=kgb.SpyOpReturn(int(time.time())))

        # First 5 should be ok.
        for i in range(5):
            result = self.api_token_auth_backend.authenticate(self.request)
            self.assertEqual(result, (False, None, None))

        # The 6th should fail.
        result = self.api_token_auth_backend.authenticate(self.request)
        self.assertEqual(
            result,
            (False, 'Maximum number of login attempts exceeded.', None))

    @override_settings(LOGIN_LIMIT_RATE='3/s')
    def test_authenticate_custom_rate_limit_3_per_second(self):
        """Testing Token Auth authenticate failed and exceeding
        number of failed login attempts allowed with custom rate limit
        (3 failed attempts per second)
        """
        token = 'myRateLimitToken1'
        self.user = User.objects.create_user(username='testratelimit1')
        MyTestWebAPITokenModel.objects.create(user=self.user, token=token)
        self.request.user = self.user
        self.request.META['HTTP_AUTHORIZATION'] = 'token bad_token'

        # Make sure that the time does not change during this test.
        self.spy_on(_get_time_int, op=kgb.SpyOpReturn(int(time.time())))

        # First 3 should be ok.
        for i in range(3):
            result = self.api_token_auth_backend.authenticate(self.request)
            self.assertEqual(result, (False, None, None))

        # The 4th should fail.
        result = self.api_token_auth_backend.authenticate(self.request)
        self.assertEqual(
            result,
            (False, 'Maximum number of login attempts exceeded.', None))

    @override_settings(LOGIN_LIMIT_RATE='4/h')
    def test_authenticate_custom_rate_limit_4_per_hour(self):
        """Testing Token Auth authenticate failed and exceeding
        number of failed login attempts allowed with custom rate limit
        (4 failed attempts per hour)
        """
        token = 'myRateLimitToken2'
        self.user = User.objects.create_user(username='testratelimit2')
        MyTestWebAPITokenModel.objects.create(user=self.user, token=token)
        self.request.user = self.user
        self.request.META['HTTP_AUTHORIZATION'] = 'token bad_token'

        # Make sure that the time does not change during this test.
        self.spy_on(_get_time_int, op=kgb.SpyOpReturn(int(time.time())))

        # First 4 should be ok.
        for i in range(4):
            result = self.api_token_auth_backend.authenticate(self.request)
            self.assertEqual(result, (False, None, None))

        # 5th should fail.
        result = self.api_token_auth_backend.authenticate(self.request)
        self.assertEqual(
            result,
            (False, 'Maximum number of login attempts exceeded.', None))

    def test_authenticate_invalid_token(self):
        """Testing Token Auth authenticate fails when using an invalid token"""
        logger = logging.getLogger('djblets.webapi.auth.backends.api_tokens')
        self.spy_on(logger.debug)

        token = 'token123'
        self.user = User.objects.create_user(username='testuser')
        invalid_date = datetime.datetime(2022, 8, 2, 5, 45)

        MyTestWebAPITokenModel.objects.create(
            user=self.user,
            token=token,
            valid=False,
            invalid_date=timezone.make_aware(invalid_date))
        self.request.user = User()
        self.request.META['HTTP_AUTHORIZATION'] = 'token %s' % token

        result = self.api_token_auth_backend.authenticate(self.request)
        debug_message = logger.debug.spy.last_call.args[0]

        self.assertTrue(debug_message.endswith('The token is invalid.'))
        self.assertEqual(
            result,
            (
                False,
                (
                    'This API token became invalid on '
                    'Aug. 2, 2022, 10:45 a.m..'
                ),
                {
                    'WWW-Authenticate': 'Basic realm="Web API"',
                }
            ))

    def test_authenticate_invalid_token_with_reason(self):
        """Testing Token Auth authenticate fails when using an invalid token
        that has an invalid reason set
        """
        logger = logging.getLogger('djblets.webapi.auth.backends.api_tokens')
        self.spy_on(logger.debug)

        token = 'token123'
        self.user = User.objects.create_user(username='testuser')
        invalid_date = datetime.datetime(2022, 8, 2, 5, 45)

        token_object = MyTestWebAPITokenModel.objects.create(
            user=self.user,
            token=token,
            valid=False,
            invalid_date=timezone.make_aware(invalid_date),
            invalid_reason='Revoked.')
        self.request.user = User()
        self.request.META['HTTP_AUTHORIZATION'] = 'token %s' % token

        result = self.api_token_auth_backend.authenticate(self.request)
        debug_message = logger.debug.spy.last_call.args[0]

        self.assertTrue(debug_message.endswith('The token is invalid.'))
        self.assertEqual(
            result,
            (
                False,
                (
                    'This API token became invalid on '
                    'Aug. 2, 2022, 10:45 a.m.: %s'
                    % token_object.invalid_reason
                ),
                {
                    'WWW-Authenticate': 'Basic realm="Web API"',
                }
            ))

    def test_authenticate_expired_token(self):
        """Testing Token Auth authenticate fails when using an expired token"""
        logger = logging.getLogger('djblets.webapi.auth.backends.api_tokens')
        self.spy_on(logger.debug)

        token = 'token123'
        self.user = User.objects.create_user(username='testuser')
        invalid_date = datetime.datetime(2022, 8, 2, 5, 45)

        MyTestWebAPITokenModel.objects.create(
            user=self.user,
            token=token,
            expires=timezone.make_aware(invalid_date))
        self.request.user = User()
        self.request.META['HTTP_AUTHORIZATION'] = 'token %s' % token

        result = self.api_token_auth_backend.authenticate(self.request)
        debug_message = logger.debug.spy.last_call.args[0]

        self.assertTrue(debug_message.endswith('The token is expired.'))
        self.assertEqual(
            result,
            (
                False,
                (
                    'This API token expired on Aug. 2, 2022, 10:45 a.m..'
                ),
                {
                    'WWW-Authenticate': 'Basic realm="Web API"',
                }
            ))

    @override_settings(AUTHENTICATION_BACKENDS=(
        'djblets.webapi.tests.test_api_auth_backend.MyTestTokenAuthBackend',
        'djblets.webapi.auth.backends.oauth2_tokens.OAuth2TokenBackendMixin',
    ))
    def test_authenticate_valid_credentials_with_multiple_backends(self):
        """Testing Token Auth authenticate succeeds with valid credentials
        and having multiple authentication backends including one that does
        not implement a validate_token method
        """
        self.spy_on(timezone.now, op=kgb.SpyOpReturn(timezone.now()))

        token = 'token123'
        self.user = User.objects.create_user(username='testuser')
        token_object = MyTestWebAPITokenModel.objects.create(user=self.user,
                                                             token=token)
        self.request.user = User()
        self.request.META['HTTP_AUTHORIZATION'] = 'token %s' % token

        result = self.api_token_auth_backend.authenticate(self.request)
        token_object.refresh_from_db(fields=('last_used',))

        self.assertEqual(result, (True, None, None))
        self.assertNotEqual(self.request.META['CSRF_COOKIE'], '')
        self.assertEqual(token_object.last_used, timezone.now())

    def test_authenticate_deprecated_token(self):
        """Testing Token Auth authenticate succeeds and returns a deprecation
        notice in a response header when using a deprecated token
        """
        token = 'token123'
        self.user = User.objects.create_user(username='testuser')

        MyTestWebAPITokenModel.objects.create(
            user=self.user,
            token=token,
            token_generator_id='legacy_sha1')
        self.request.user = User()
        self.request.META['HTTP_AUTHORIZATION'] = 'token %s' % token

        result = self.api_token_auth_backend.authenticate(self.request)

        self.assertEqual(
            result,
            (
                True,
                None,
                {
                    'X-API-Token-Deprecated': (
                        'This token uses a deprecated format. The token can '
                        'still be used, but you should remove it and '
                        'generate a new one to take advantage of security '
                        'improvements.'
                    ),
                }
            ))



"""Unit tests for djblets.privacy.consent.forms.ConsentRequirementWidget."""

from djblets.privacy.consent import BaseConsentRequirement
from djblets.privacy.consent.forms import ConsentRequirementWidget
from djblets.privacy.tests.testcases import ConsentTestCase


class MyConsentRequirement(BaseConsentRequirement):
    requirement_id = 'my-requirement'
    name = 'My Requirement'
    summary = 'We would like to use this thing'
    intent_description = 'We need this for testing.'
    data_use_description = 'Sending all the things.'
    icons = {
        '1x': '/static/consent.png',
        '2x': '/static/consent@2x.png',
    }


class ConsentRequirementWidgetTests(ConsentTestCase):
    """Unit tests for ConsentRequirementWidget."""

    def setUp(self):
        super(ConsentRequirementWidgetTests, self).setUp()

        self.widget = ConsentRequirementWidget(
            consent_requirement=MyConsentRequirement())

    def test_render(self):
        """Testing ConsentRequirementWidget.render with value=None"""
        html = self.widget.render(
            name='consent',
            value=None,
            attrs={
                'id': 'my_consent',
            })

        self.assertIn('<div id="my_consent" class="privacy-consent-field'
                      ' privacy-consent-field-has-icon">',
                      html)
        self.assertIn("<h2>We would like to use this thing</h2>", html)
        self.assertIn('<p>We need this for testing.</p>', html)
        self.assertIn('<p>Sending all the things.</p>', html)
        self.assertInHTML('<input type="radio" id="my_consent_choice_allow"'
                          ' name="consent_choice" value="allow">',
                          html)
        self.assertInHTML('<input type="radio" id="my_consent_choice_block"'
                          ' name="consent_choice" value="block">',
                          html)

    def test_render_with_value_allow(self):
        """Testing ConsentRequirementWidget.render with value=allow"""
        html = self.widget.render(
            name='consent',
            value='allow',
            attrs={
                'id': 'my_consent',
            })

        self.assertIn('<div id="my_consent" class="privacy-consent-field'
                      ' privacy-consent-field-has-icon">',
                      html)
        self.assertIn("<h2>We would like to use this thing</h2>", html)
        self.assertIn('<p>We need this for testing.</p>', html)
        self.assertIn('<p>Sending all the things.</p>', html)
        self.assertInHTML('<input type="radio" id="my_consent_choice_allow"'
                          ' name="consent_choice" value="allow" checked>',
                          html)
        self.assertInHTML('<input type="radio" id="my_consent_choice_block"'
                          ' name="consent_choice" value="block">',
                          html)

    def test_render_with_value_block(self):
        """Testing ConsentRequirementWidget.render with value=block"""
        html = self.widget.render(
            name='consent',
            value='block',
            attrs={
                'id': 'my_consent',
            })

        self.assertIn('<div id="my_consent" class="privacy-consent-field'
                      ' privacy-consent-field-has-icon">',
                      html)
        self.assertIn("<h2>We would like to use this thing</h2>", html)
        self.assertIn('<p>We need this for testing.</p>', html)
        self.assertIn('<p>Sending all the things.</p>', html)
        self.assertInHTML('<input type="radio" id="my_consent_choice_allow"'
                          ' name="consent_choice" value="allow">',
                          html)
        self.assertInHTML('<input type="radio" id="my_consent_choice_block"'
                          ' name="consent_choice" value="block" checked>',
                          html)

    def test_value_from_datadict_with_value(self):
        """Testing ConsentRequirementWidget.value_from_datadict with value"""
        data = {
            'consent_choice': 'allow',
        }

        self.assertEqual(
            self.widget.value_from_datadict(data=data,
                                            files={},
                                            name='consent'),
            'allow')

    def test_value_from_datadict_with_none(self):
        """Testing ConsentRequirementWidget.value_from_datadict without value
        """
        self.assertIsNone(
            self.widget.value_from_datadict(data={}, files={}, name='consent'))



"""Utility mixins for registries."""

from typing import Generic, Optional, TYPE_CHECKING

from djblets.registries.errors import ItemLookupError
from djblets.registries.registry import RegistryItemType


if TYPE_CHECKING:
    # When checking types, we set the base class to a custom class that has
    # a compatible function signature. This is done for two reasons:
    #
    # 1. So that the type checker doesn't see a call to super().get() as
    #    being an invalid method on Generic.
    #
    # 2. So the get() method has a compatible return type and doesn't result
    #    in errors during type checking (which would happen if subclassing
    #    Registry or using a Protocol).
    class _BaseClass(Generic[RegistryItemType]):
        def get(
            self,
            attr_name: str,
            attr_value: object,
        ) -> Optional[RegistryItemType]:
            ...
else:
    _BaseClass = Generic


class ExceptionFreeGetterMixin(_BaseClass[RegistryItemType]):
    """A mixin that prevents lookups from throwing errors.

    If using typed registries and inheriting from this mixin, the same type
    should be passed to this class as well.

    This class may be deprecated in the future. Callers may want to switch
    to calling :py:meth:`Registry.get_or_none()
    <djblets.registries.registry.Registry.get_or_none>` instead, which offers
    the same functionality.

    Version Changed:
        3.1:
        Added support for specifying a registry item type when subclassing
        this mixin.
    """

    def get(
        self,
        attr_name: str,
        attr_value: object,
    ) -> Optional[RegistryItemType]:
        """Return the requested registered item.

        Args:
            attr_name (str):
                The attribute name.

            attr_value (object):
                The attribute value.

        Returns:
            object:
            The matching registered item, if found. Otherwise, ``None`` is
            returned.
        """
        # This effectively duplicates Registry.get_or_none(), but since that
        # calls get(), we can't use it or we'd create an infinite loop.
        try:
            return super().get(attr_name, attr_value)
        except ItemLookupError:
            return None



"""Unit tests for djblets.extensions.hooks.BaseRegistryMultiItemsHook."""

from djblets.extensions.extension import Extension
from djblets.extensions.hooks import (ExtensionHookPoint,
                                      BaseRegistryMultiItemHook)
from djblets.extensions.testing import ExtensionTestCaseMixin
from djblets.registries.errors import AlreadyRegisteredError
from djblets.registries.registry import Registry
from djblets.testing.testcases import TestCase


class DummyRegistry(Registry):
    lookup_attrs = ('foo_id',)


class DummyItem(object):
    def __init__(self, foo_id):
        self.foo_id = foo_id


class MyTestExtension(Extension):
    pass


class BaseRegistryMultiItemHookTests(ExtensionTestCaseMixin, TestCase):
    """Unit tests for djblets.extensions.hooks.BaseRegistryMultiItemsHook."""

    extension_class = MyTestExtension

    def setUp(self):
        super(BaseRegistryMultiItemHookTests, self).setUp()

        self.registry = DummyRegistry()

        class DummyRegistryHook(BaseRegistryMultiItemHook,
                                metaclass=ExtensionHookPoint):
            registry = self.registry

        self.hook_cls = DummyRegistryHook

    def test_initialize(self):
        """Testing BaseRegistryMultiItemHook.initialize"""
        self.assertEqual(list(self.registry), [])

        item1 = DummyItem(123)
        item2 = DummyItem(456)
        self.hook_cls(self.extension, [item1, item2])

        self.assertIn(item1, self.registry)
        self.assertIn(item2, self.registry)

    def test_initialize_handles_errors(self):
        """Testing BaseRegistryMultiItemHook.initialize unregisters all items
        on registration error
        """
        self.assertEqual(list(self.registry), [])

        item1 = DummyItem(123)
        item2 = DummyItem(123)

        with self.assertRaises(AlreadyRegisteredError):
            self.hook_cls(self.extension, [item1, item2])

        self.assertEqual(list(self.registry), [])

    def test_shutdown(self):
        """Testing BaseRegistryMultiItemHook.shutdown"""
        self.assertEqual(list(self.registry), [])

        item1 = DummyItem(123)
        item2 = DummyItem(456)
        self.hook_cls(self.extension, [item1, item2])

        self.extension.shutdown()
        self.assertEqual(list(self.registry), [])



"""Markdown extension to render content similar to the source in e-mails.

This will render Markdown such that the spacing and alignment in the source
text and the rendered content looks roughly the same. It's meant to help ensure
that what's typed is very close to what's viewed when rendered.
"""

from django import template
from markdown.extensions import Extension
from markdown.treeprocessors import Treeprocessor


register = template.Library()


class InlineStyleProcessor(Treeprocessor):
    """Injects CSS styles directly into the tags, for use in e-mails.

    This will process each element and, depending on the element type,
    create inline styles. These styles are meant to match the stylesheets
    in the :file:`markdown-wysiwyg.less` file.
    """

    def process_element(self, context, el):
        """Process an element in the tree.

        This adds a handful of inline styles to the resulting document which
        mimic the .rich-text rules in markdown-wysiwyg.less.

        This does not do quite everything that the markdown-wysiwyg.less rules
        do, due to the complexity in implementing CSS selectors in Python,
        and the CSS compatibility logic in most e-mail clients.

        The end result is that the e-mail will look similar but not identical
        to the page.
        """
        style = {
            'margin': 0,
            'padding': 0,
            'line-height': 'inherit',
            'text-rendering': 'inherit',
            'white-space': 'normal',
        }
        tag = el.tag

        if tag in ('h1', 'h2', 'h3', 'h4', 'h5', 'h6'):
            style['font-size'] = '100%'
        elif tag == 'ul':
            style['margin'] = '0 0 0 1em'
        elif tag == 'ol':
            style['margin'] = '0 0 0 2em'
        elif tag == 'code':
            style['color'] = '#4444cc'
        elif tag == 'p':
            style['white-space'] = 'inherit'
        elif tag == 'blockquote':
            style.update({
                'border-left': '1px solid #bbb',
                'padding': '0 0 0 1em',
                'margin': '0 0 0 0.5em',
            })
        elif tag == 'hr':
            style['border'] = '1px solid #ddd'
        elif tag in ('th', 'td'):
            style.update({
                'border-bottom': '1px solid #bbb',
                'padding': '0.2em 1em',
            })

        # Create a valid CSS string and set it as the style attribute
        el.set('style', ''.join(
            '%s: %s;' % (k, v)
            for k, v in style.items()
        ))

        # Recurse into children
        context.append(tag)

        for child in el:
            self.process_element(context, child)

        context.pop()

    def run(self, root):
        # We don't just process the root element, because if we add any style
        # characteristics to it, markdown can't strip off the top-level <div>
        # tag. Why it doesn't use the element tree to do the stripping is
        # beyond me.
        for child in root:
            self.process_element([], child)


class WysiwygEMailExtension(Extension):
    """A Markdown extension to inject CSS into elements for HTML output.

    This is meant as a counterpart to the Wysiwyg Markdown extension. It
    aims to provide HTML output that looks as similar as possible to the
    input, preserving the spacing, font sizes, alignment, and other styles
    from the raw text.

    This is meant to be used with the following Markdown configuration
    and extensions:

    .. code-block:: python

        {
            'extensions': [
                'codehilite(noclasss=True)', 'tables',
                'djblets.markdown.extentions.wysiwyg',
            ],
        }
    """

    def extendMarkdown(self, md):
        """Extend the list of Markdown processors.

        Args:
            md (markdown.Markdown):
                The Markdown instance.
        """
        md.treeprocessors.register(
            InlineStyleProcessor(),
            'inlinestyle',
            priority=0)


def makeExtension(*args, **kwargs):
    """Create and return an instance of this extension.

    Args:
        *args (tuple):
            Positional arguments for the extension.

        **kwargs (dict):
            Keyword arguments for the extension.

    Returns:
        WysiwygEMailExtension:
        The extension instance.
    """
    return WysiwygEMailExtension(*args, **kwargs)



import io
import json
from xml.sax.saxutils import XMLGenerator

from django.conf import settings
from django.contrib.auth.models import User, Group
from django.db.models.query import QuerySet

from djblets.util.serializers import DjbletsJSONEncoder


class WebAPIEncoder(object):
    """Encodes an object into a dictionary of fields and values.

    This object is used for both JSON and XML API formats.

    Projects can subclass this to provide representations of their objects.
    To make use of a encoder, add the path to the encoder class to
    the project's ``settings.WEB_API_ENCODERS`` list.

    For example:

    .. code-block:: python

        WEB_API_ENCODERS = (
            'myproject.webapi.MyEncoder',
        )
    """

    def encode(self, o, *args, **kwargs):
        """Encodes an object.

        This is expected to return either a dictionary or a list. If the
        object being encoded is not supported, return None, or call
        the superclass's encode method.
        """
        return None


class BasicAPIEncoder(WebAPIEncoder):
    """A basic encoder that encodes standard types.

    This supports encoding of dates, times, QuerySets, Users, and Groups.
    """
    def encode(self, o, *args, **kwargs):
        if isinstance(o, QuerySet):
            return list(o)
        elif isinstance(o, User):
            return {
                'id': o.id,
                'username': o.username,
                'first_name': o.first_name,
                'last_name': o.last_name,
                'fullname': o.get_full_name(),
                'email': o.email,
                'url': o.get_absolute_url(),
            }
        elif isinstance(o, Group):
            return {
                'id': o.id,
                'name': o.name,
            }
        else:
            try:
                return DjbletsJSONEncoder().default(o)
            except TypeError:
                return None


class ResourceAPIEncoder(WebAPIEncoder):
    """An encoder that encodes objects based on registered resources."""
    def encode(self, o, *args, **kwargs):
        if isinstance(o, QuerySet):
            return list(o)
        else:
            calling_resource = kwargs.pop('calling_resource', None)

            if calling_resource:
                serializer = calling_resource.get_serializer_for_object(o)
            else:
                from djblets.webapi.resources import get_resource_for_object

                serializer = get_resource_for_object(o)

            if serializer:
                return serializer.serialize_object(o, *args, **kwargs)
            else:
                try:
                    return DjbletsJSONEncoder().default(o)
                except TypeError:
                    return None


class JSONEncoderAdapter(json.JSONEncoder):
    """Adapts a WebAPIEncoder to be used with json.

    This takes an existing encoder and makes it available to use as a
    json.JSONEncoder. This is used internally when generating JSON from a
    WebAPIEncoder, but can be used in other projects for more specific
    purposes as well.
    """
    def __init__(self, encoder, *args, **kwargs):
        super(JSONEncoderAdapter, self).__init__(
            sort_keys=kwargs.pop('sort_keys', True),
            *args, **kwargs)
        self.encoder = encoder

    def encode(self, o, *args, **kwargs):
        self.encode_args = args
        self.encode_kwargs = kwargs
        return super(JSONEncoderAdapter, self).encode(o)

    def default(self, o):
        """Encodes an object using the supplied WebAPIEncoder.

        If the encoder is unable to encode this object, a TypeError is raised.
        """
        result = self.encoder.encode(o, *self.encode_args,
                                     **self.encode_kwargs)

        if result is None:
            raise TypeError("%r is not JSON serializable" % (o,))

        return result


class XMLEncoderAdapter(object):
    """Adapts a WebAPIEncoder to output XML.

    This takes an existing encoder and adapts it to output a simple XML format.
    """
    def __init__(self, encoder, *args, **kwargs):
        self.encoder = encoder

    def encode(self, o, *args, **kwargs):
        self.level = 0
        self.doIndent = False

        stream = io.StringIO()
        self.xml = XMLGenerator(stream, settings.DEFAULT_CHARSET)
        self.xml.startDocument()
        self.startElement("rsp")
        self.__encode(o, *args, **kwargs)
        self.endElement("rsp")
        self.xml.endDocument()
        self.xml = None

        return stream.getvalue()

    def __encode(self, o, *args, **kwargs):
        if isinstance(o, dict):
            for key, value in o.items():
                attrs = {}

                if isinstance(key, int):
                    attrs['value'] = str(key)
                    key = 'int'

                self.startElement(key, attrs)
                self.__encode(value, *args, **kwargs)
                self.endElement(key)
        elif isinstance(o, (tuple, list)):
            self.startElement("array")

            for i in o:
                self.startElement("item")
                self.__encode(i, *args, **kwargs)
                self.endElement("item")

            self.endElement("array")
        elif isinstance(o, str):
            self.text(o)
        elif isinstance(o, int):
            self.text("%d" % o)
        elif isinstance(o, float):
            self.text("%s" % o)
        elif isinstance(o, bool):
            if o:
                self.text("True")
            else:
                self.text("False")
        elif o is None:
            pass
        else:
            result = self.encoder.encode(o, *args, **kwargs)

            if result is None:
                raise TypeError("%r is not XML serializable" % (o,))

            return self.__encode(result, *args, **kwargs)

    def startElement(self, name, attrs={}):
        self.addIndent()
        self.xml.startElement(name, attrs)
        self.level += 1
        self.doIndent = True

    def endElement(self, name):
        self.level -= 1
        self.addIndent()
        self.xml.endElement(name)
        self.doIndent = True

    def text(self, value):
        self.xml.characters(value)
        self.doIndent = False

    def addIndent(self):
        if self.doIndent:
            self.xml.ignorableWhitespace('\n' + ' ' * self.level)


_registered_encoders = None


def get_registered_encoders():
    """
    Returns a list of registered Web API encoders.
    """
    global _registered_encoders

    if _registered_encoders is None:
        _registered_encoders = []

        encoders = getattr(settings, 'WEB_API_ENCODERS',
                           ['djblets.webapi.encoders.BasicAPIEncoder'])

        for encoder in encoders:
            encoder_path = encoder.split('.')
            if len(encoder_path) > 1:
                encoder_module_name = '.'.join(encoder_path[:-1])
            else:
                encoder_module_name = '.'

            encoder_module = __import__(encoder_module_name, {}, {},
                                        encoder_path[-1])
            encoder_class = getattr(encoder_module, encoder_path[-1])
            _registered_encoders.append(encoder_class())

    return _registered_encoders



"""Exception classes for dealing with Djblets registries."""


class ItemLookupError(Exception):
    """An error that occurs during item lookup."""


class RegistrationError(Exception):
    """An error that occurs during registration."""


class AlreadyRegisteredError(RegistrationError):
    """An error that occurs during registering the same item."""
